#include "DapExpConvert.h"
#include <QRegularExpression>
#include <QString>
#include <QRegExp>
#include "../DapTypes/DapCoin.h"

DapExpConvert::DapExpConvert(QObject *parent) : QObject(parent)
{

}

bool DapExpConvert::isExp(const QString& str)
{
    QStringList eSplit = str.split('e');
    if(eSplit.count() != 2) return false;

    QString mantissa = eSplit[0];
    QString exponent = eSplit[1];

    // check exponent
    QChar digit = exponent[0];
    QString expSize = exponent.mid(1);

    if(!(digit == '+' || digit == '-')) return false;
    if(!containsOnlyDigits(expSize)) return false;

    // check mantissa
    if(mantissa.indexOf('.') < 0)
    {
        if(!containsOnlyDigits(mantissa)) return false;
    }
    else
    {
        QStringList dotSplit = mantissa.split('.');
        if(dotSplit.count() > 2) return false;
        if(!containsOnlyDigits(mantissa.remove('.'))) return false;
    }
    return true;
}

bool DapExpConvert::isDatoshi(const QString &str)
{
    const QRegularExpression regular = QRegularExpression(R"(^[1-9][0-9]*$)");
    return str.contains(regular);
}

QString DapExpConvert::expToFixed(const QString& src)
{
    if(!isExp(src)) return src;

    // Find matissa & exponent
    QStringList eSplit = src.split('e');

    QString mantissa = eSplit[0];
    QString exponent = eSplit[1];

    int sign = exponent[0] == '+' ? +1 : exponent[0] == '-' ? -1 : 0;
    int expSize = exponent.mid(1).toInt();

    int dotLeftCount;
    int dotRightCount = 0;
    if(mantissa.indexOf('.') < 0)
    {
        dotLeftCount = mantissa.length();
    }
    else
    {
        QStringList dotSplit = mantissa.split('.');
        dotLeftCount = dotSplit[0].length();
        dotRightCount = dotSplit[1].length();
    }

    // Generate result
    QString result = mantissa;
    if(sign > 0)
    {
        result = result.remove('.');
        if(dotRightCount > expSize)
        {
            result.insert(dotLeftCount + expSize, '.');
        }
        else
        {
            for(int i = 0; i < expSize - dotRightCount; i++) result.append('0');
            result = result + ".0";
        }
    }
    else
    {
        result = result.remove('.');
        if(dotLeftCount > expSize)
        {
            result.insert(dotLeftCount - expSize, '.');
        }
        else
        {
            for(int i = 0; i < expSize - dotLeftCount; i++) result.insert(0, '0');
            result.insert(0, "0.");
        }
    }

    result = removeNulls(result);
    result = setLimitSymbols(result, 18);

    //qDebug() << "expToFixed() Source:" << src << "  Result:" << result;

    return result;
}

QString DapExpConvert::datoshiToCoins(const QString& datoshi)
{
    Dap::Coin result;
    result.fromDatoshi(datoshi);
    return result.toCoinsString();
}

QString DapExpConvert::coinsToDatoshi(const QString& coins)
{
    const int exponent = 18;

    QString src = DapExpConvert::expToFixed(coins);

    QStringList dotSplit = src.split('.');
    int dotLeftCount = dotSplit.count() == 2 ? dotSplit[0].length() : src.length();
    int dotRightCount = dotSplit.count() == 2 ? dotSplit[1].length() : 0;

    QString result = src.remove('.');
    if(dotRightCount > exponent)
    {
        result.insert(dotLeftCount + exponent, '.');
    }
    else
    {
        for(int i = 0; i < exponent - dotRightCount; i++) result.append('0');
        result = result + ".0";
    }

    result = removeNulls(result);
    result = setLimitSymbols(result, 18);

    //qDebug() << "coinsToDatoshi(" << exponent << ") Source:" << coins << "  Result:" << result;

    return result;
}

QString DapExpConvert::removeNulls(const QString &src)
{
    QString result = src;

    // front
    while(result[0] == '0')
    {
        if(result[1] == '.') break;
        result.remove(0,1);
    }

    // end
    QStringList dotSplit = result.split('.');
    if(dotSplit.count() == 2)
    {
        while(result[result.length()-1] == '0')
        {
            result.remove(result.length()-1,1);
        }
        if(result[result.length()-1] == '.') result.remove(result.length()-1,1);
    }

    return result;
}

QString DapExpConvert::setLimitSymbols(const QString &src, int count)
{
    QString result = src;

    QStringList dotSplit = result.split('.');
    if(dotSplit.count() == 2)
    {
        if(dotSplit[1].length() > count)
        {
            result = dotSplit[0] + '.' + dotSplit[1].left(count);
        }
    }

    return result;
}

bool DapExpConvert::containsOnlyDigits(const QString& str)
{
    QRegularExpression regex("^[0-9]+$");
    QRegularExpressionMatch match = regex.match(str);
    return match.hasMatch();
}
