#include "DapConfigToolController.h"

DapConfigToolController::DapConfigToolController(QObject *parent)
    : QObject{parent}
{
    m_initTimer = new QTimer(this);

    if(initConfTool())
        return;
    else
    {
        m_initTimer->start(1000);
        connect(m_initTimer, &QTimer::timeout, this, [this] {
            qDebug()<<"Reinit timer tick";
            if(initConfTool())
                m_initTimer->stop();
        });
    }
}

DapConfigToolController &DapConfigToolController::getInstance()
{
    static DapConfigToolController instance;
    return instance;
}

bool DapConfigToolController::runNode()
{
    if(!m_statusInitConfTool)
    {
        qWarning()<<"ConfigToolController is not init";
        return false;
    }

    if(m_flagUserStopAutostart)
        return false;

    if(!serviceCommand(Status)["result"].toString().contains("enabled"));
        serviceCommand(EnableAutostart);
    return serviceCommand(Start)["result"].toString().contains("started");
}

void DapConfigToolController::stopNode()
{
    serviceCommand(Stop);
}

void DapConfigToolController::startNode()
{
    serviceCommand(Start);
}

void DapConfigToolController::getStatusNode()
{
    serviceCommand(Status);
}

void DapConfigToolController::swithServiceEnabled(bool flag)
{
    m_flagUserStopAutostart = !flag;
    serviceCommand(flag ? EnableAutostart : DisableAutostart);
}

void DapConfigToolController::initPathConftool()
{
#if defined(Q_OS_LINUX) && !defined(Q_OS_ANDROID)
    m_nodeConfToolPath = "/opt/cellframe-node/bin/cellframe-node-config";
#elif defined (Q_OS_MACOS)
    m_nodeConfToolPath = "/Applications/CellframeNode.app/Contents/MacOS/cellframe-node-config";
#elif defined (Q_OS_WIN)
    //HKLM "Software\${APP_NAME}" "Path"
    HKEY hKey;
    LONG lRes = RegOpenKeyExW(HKEY_LOCAL_MACHINE, L"SOFTWARE\\cellframe-node\\", 0, KEY_READ, &hKey);
    bool bExistsAndSuccess (lRes == ERROR_SUCCESS);
    bool bDoesNotExistsSpecifically (lRes == ERROR_FILE_NOT_FOUND);
    std::wstring path;
    GetStringRegKey(hKey, L"Path", path, L"");
    std::string stdpath(path.begin(),path.end());

    m_nodeConfToolPath = QString(QString::fromWCharArray(path.c_str()) + "/cellframe-node-config.exe");
#endif

    m_statusInitConfTool = checkConftool();
    emit changeStatusConftoolInit(m_statusInitConfTool);
}

bool DapConfigToolController::checkConftool()
{
    QFileInfo confTool(m_nodeConfToolPath);
    if(!confTool.exists())
    {
        qWarning()<<"[DapConfigToolController] Could not find cellframe-node-config";
        return false;
    }
    qInfo()<<"[DapConfigToolController] Find cellframe-node-config";
    qInfo()<<"[DapConfigToolController] ConfTool path: " << m_nodeConfToolPath;
    return true;
}

bool DapConfigToolController::initConfTool()
{
    initPathConftool();

    if(m_statusInitConfTool)
    {
        serviceCommand(Status);

        return true;
    }
    return false;
}

QString DapConfigToolController::getResult(QString find, QStringList list)
{
    for(int i = 0; i < list.length(); i++)
    {
        if(list[i].contains(find))
            return list[i].remove(find);
    }
    return "";
}

QStringList DapConfigToolController::getConfigNetworkList(const QString& status)
{
    QStringList netList;

    QString result = sendRequest("-e net_list " + status);
    auto list = result.split('\n');
    const QRegularExpression regName = QRegularExpression(R"(name: (\S+))");

    for(auto& str: list)
    {
        str = str.trimmed();
        if(str.isEmpty())continue;

        auto nameMatch = regName.match(str);
        if(!nameMatch.hasMatch())
        {
            qWarning() << "[DapConfigToolController] Problems getting a list of networks. >> " << result;
            continue;
        }

        netList.append(nameMatch.captured(1));
    }

    return netList;
}

void DapConfigToolController::setConfigParam(const QString& configName, const QString& block, const QString& param, const QString& value)
{
    QString command = QString("-e config %1 %2 %3 ensure %4").arg(configName).arg(block).arg(param).arg(value);
    QString result = sendRequest(command);
    qDebug() << "[DapConfigToolController] [setConfigParam] command: " << command;
    qDebug() << "[DapConfigToolController] [setConfigParam] result: " << result;
}

QString DapConfigToolController::getConfigParam(const QString &configName, const QString &block, const QString &param)
{
    QString command = QString("-e config %1 %2 %3 get").arg(configName).arg(block).arg(param);
    QString result = sendRequest(command);

    result = result.split(param).last();
    result.remove("=").remove(" ").remove("\n").remove("\t").remove("\r");

//    qDebug() << "[DapConfigToolController] [getConfigParam] command: " << command;
//    qDebug() << "[DapConfigToolController] [getConfigParam] result: " << result;
    return result;
}

QStringList DapConfigToolController::parseArrayConfigParam(QString str)
{

    auto isValidInput = [](const QString& input) -> bool
    {
        QRegularExpression regex(R"(^\[\s*([^,\[\]]+\s*(,\s*[^,\[\]]+\s*)*)?\]$)");
        return regex.match(input).hasMatch();
    };

    if (!isValidInput(str)) {
//        qDebug() << "Error: Unknown input text format. " << str;
        return QStringList()<<str;
    }

    str.remove(0, 1); // remove '['
    str.remove(str.length() - 1, 1); // remove ']'

    QStringList variables = str.split(",");
    QStringList result;

    for (const QString &var : qAsConst(variables)) {
        result.append(var.trimmed().remove(".."));
    }

    return result;
}

QJsonObject DapConfigToolController::serviceCommand(TypeServiceCommands type)
{
    QString result;
    bool status;

    switch (type) {
    case Status:
    {
        result = sendRequest("-e service status");

        QStringList resList = result.split("\n");
        m_statusProcessNode = getResult("process: ", resList).contains("running");
        m_statusServiceNode = getResult("service: ", resList).contains("enabled");

        emit statusProcessNodeChanged();
        emit statusServiceNodeChanged();
        break;
    }
    case EnableAutostart:
    {
        result = sendRequest("-e service enable");
        serviceCommand(Status);
        break;
    }
    case DisableAutostart:
    {
        result = sendRequest("-e service disable");
        serviceCommand(Status);
        break;
    }
    case Stop:
    {
        result = sendRequest("-e service stop");
        serviceCommand(Status);
        break;
    }
    case Start:
    {
        result = sendRequest("-e service start");
        serviceCommand(Status);
        break;
    }
    case Restart:
    {
        result = sendRequest("-e service restart");
        serviceCommand(Status);
        break;
    }
    default:
        break;
    }

    qDebug()<< result;

    status = !result.isEmpty();
    return QJsonObject{{"status", status},
                       {"result", result}};
}

QString DapConfigToolController::sendRequest(QString req)
{
    QProcess proc;
    proc.setProgram(m_nodeConfToolPath);
    proc.setArguments(req.split(" "));

    proc.start();

    if(proc.waitForFinished())
    {
        return proc.readAll();
    }

    qDebug()<< "Timeout request";
    return QString();
}

#ifdef WIN32
LONG DapConfigToolController::GetDWORDRegKey(HKEY hKey, const std::wstring &strValueName, DWORD &nValue, DWORD nDefaultValue)
{
    nValue = nDefaultValue;
    DWORD dwBufferSize(sizeof(DWORD));
    DWORD nResult(0);
    LONG nError = ::RegQueryValueExW(hKey,
                                     strValueName.c_str(),
                                     0,
                                     NULL,
                                     reinterpret_cast<LPBYTE>(&nResult),
                                     &dwBufferSize);
    if (ERROR_SUCCESS == nError)
    {
        nValue = nResult;
    }
    return nError;
}


LONG DapConfigToolController::GetBoolRegKey(HKEY hKey, const std::wstring &strValueName, bool &bValue, bool bDefaultValue)
{
    DWORD nDefValue((bDefaultValue) ? 1 : 0);
    DWORD nResult(nDefValue);
    LONG nError = GetDWORDRegKey(hKey, strValueName.c_str(), nResult, nDefValue);
    if (ERROR_SUCCESS == nError)
    {
        bValue = (nResult != 0) ? true : false;
    }
    return nError;
}


LONG DapConfigToolController::GetStringRegKey(HKEY hKey, const std::wstring &strValueName, std::wstring &strValue, const std::wstring &strDefaultValue)
{
    strValue = strDefaultValue;
    WCHAR szBuffer[512];
    DWORD dwBufferSize = sizeof(szBuffer);
    ULONG nError;
    nError = RegQueryValueExW(hKey, strValueName.c_str(), 0, NULL, (LPBYTE)szBuffer, &dwBufferSize);
    if (ERROR_SUCCESS == nError)
    {
        strValue = szBuffer;
    }
    return nError;
}
#endif
