#include "DapNodePathManager.h"

DapNodePathManager::DapNodePathManager( QObject *parent)
    : QObject(parent)
    , m_instMngr(new DapNodeInstallManager(true))
    , m_cfgToolCtrl(&DapConfigToolController::getInstance())
    , m_sharedMemory(m_keyName)
{
    m_target = QString(PROG_TYPE);

    connect(m_instMngr, &DapNodeInstallManager::singnalReadyUpdateToNode, this, [this] (bool isReady)
    {
        emit checkedUrlSignal(isReady);
    });

    connect(m_cfgToolCtrl, &DapConfigToolController::changeStatusConftoolInit, this, [this] (bool isReady)
    {
        if(isReady)
            init();
    });
    init();
}

DapNodePathManager &DapNodePathManager::getInstance()
{
    static DapNodePathManager instance;
    return instance;
}

void DapNodePathManager::init()
{
    fillNodePath();

    m_initMemFlag = initMem();

    qDebug()<<"--------------------------- Init QShareMemory ---------------------------"
            <<"Mem target: "               + m_target
            <<"Mem status: "               + QString(m_initMemFlag?"true":"false")
            <<"Node install type: "        + QString::number(nodePaths.nodeInstallType)
            <<"Mem cli path: "             + nodePaths.nodePath_cli
            <<"Mem tool path: "            + nodePaths.nodePath_tool
            <<"Mem dir path: "             + nodePaths.nodeDirPath
            <<"Mem node path: "            + nodePaths.nodePath
            <<"Mem node config path: "     + nodePaths.nodeConfigPath;

    emit installNodeChanged();
    checkNeedDownload();
}

void DapNodePathManager::slotReinit()
{
    fillNodePath();
    writeMem();

    qDebug()<<"--------------------------- Reinit QShareMemory ---------------------------"
            <<"Mem target: "               + m_target
            <<"Node install type: "        + QString::number(nodePaths.nodeInstallType)
            <<"Mem cli path: "             + nodePaths.nodePath_cli
            <<"Mem tool path: "            + nodePaths.nodePath_tool
            <<"Mem dir path: "             + nodePaths.nodeDirPath
            <<"Mem node path: "            + nodePaths.nodePath
            <<"Mem node config path: "     + nodePaths.nodeConfigPath;

    emit installNodeChanged();
    checkNeedDownload();

    emit signalReinitFinish();
}

QString DapNodePathManager::getUrlForNodeDownload()
{
    return m_instMngr->getUrlForDownload();
}

QString DapNodePathManager::getNodeUrl(const QString& ver)
{
    if(ver.isEmpty())
    {
        return getUrlForNodeDownload();
    }
    return m_instMngr->getUrl(ver);
}

void DapNodePathManager::tryCheckUrl(const QString& url)
{
    m_instMngr->checkUpdateNode(url);
}

void DapNodePathManager::checkNeedDownload()
{
    if(m_target == "GUI")
    {
        switch (nodePaths.nodeInstallType) {
        case Unknown:
            emit signalIsNeedInstallNode(false, "");
            qDebug()<<"Unknown type node install";
            break;
        case NoInstall:
            emit signalIsNeedInstallNode(true, getUrlForNodeDownload());
            break;
        default:
            emit signalIsNeedInstallNode(false, "");
            break;
        }
    }
}

void DapNodePathManager::slotCheckUpdateNode(QString currentNodeVersion)
{
    m_instMngr->checkUpdateNode(currentNodeVersion);
}

bool DapNodePathManager::initMem()
{
    if (m_sharedMemory.isAttached())
        m_sharedMemory.detach();

    return readMem();
}

bool DapNodePathManager::readMem()
{
    if (!m_sharedMemory.attach()) {
        qDebug()<<"Unable to attach to shared memory segment. " << m_sharedMemory.errorString() << ". Write attempt mem";
        return writeMem();
    }

    QBuffer buffer;
    QDataStream in(&buffer);

    m_sharedMemory.lock();
    buffer.setData((char*)m_sharedMemory.constData(), m_sharedMemory.size());
    buffer.open(QBuffer::ReadOnly);
    in >> nodePaths.nodePath
       >> nodePaths.nodeDirPath
       >> nodePaths.nodePath_cli
       >> nodePaths.nodePath_tool
       >> nodePaths.nodeInstallType;
    m_sharedMemory.unlock();

    m_sharedMemory.detach();

    return true;
}

bool DapNodePathManager::writeMem()
{
    // load into shared memory
    QBuffer buffer;
    buffer.open(QBuffer::ReadWrite);
    QDataStream out(&buffer);
    out << nodePaths.nodePath
        << nodePaths.nodeDirPath
        << nodePaths.nodePath_cli
        << nodePaths.nodePath_tool
        << nodePaths.nodeInstallType;
    int size = buffer.size();

    if (!m_sharedMemory.create(size)) {
        qDebug()<<"Unable to create shared memory segment." << m_sharedMemory.errorString();
        return false;
    }

    m_sharedMemory.lock();
    char *to = (char*)m_sharedMemory.data();
    const char *from = buffer.data().data();
    memcpy(to, from, qMin(m_sharedMemory.size(), size));
    m_sharedMemory.unlock();

    return true;
}

void DapNodePathManager::fillNodePath()
{
    if(!m_cfgToolCtrl->m_statusInitConfTool)
    {
        qWarning()<<"[DapNodePathManager] ConfTool does't init";
        nodePaths.nodeInstallType = NoInstall;
        return;
    }

    QString varsResult = m_cfgToolCtrl->sendRequest("--vars").remove("\r");

    if(varsResult.isEmpty())
    {
        qWarning()<<"Empty result --vars";
        return;
    }

    const QRegularExpression rxVars = QRegularExpression(R"(CONFIGS_PATH=(.+)\nHOSTNAME=(.+)\nHOST_OS=(.+)\nNODE_BINARY_PATH=(.+))");
    QRegularExpressionMatch matchVars = rxVars.match(varsResult);
    if(!matchVars.hasMatch())
    {
        qWarning()<<"Error parse /'--vars/'";
        return;
    }

    QString configs_path, hostname, host_os, node_binary_path;

    configs_path     = matchVars.captured(1);
    hostname         = matchVars.captured(2);
    host_os          = matchVars.captured(3);
    node_binary_path = matchVars.captured(4);

#ifdef Q_OS_WIN
    QString suffix = ".exe";
#else
    QString suffix = "";
#endif

    nodePaths.nodePath        = node_binary_path + "/" + "cellframe-node" + suffix;
    nodePaths.nodePath_cli    = node_binary_path + "/" + "cellframe-node-cli" + suffix;
    nodePaths.nodePath_tool   = node_binary_path + "/" + "cellframe-node-tool" + suffix;
    nodePaths.nodeDirPath     = node_binary_path.remove("bin/");
    nodePaths.nodeConfigPath  = configs_path;
    nodePaths.nodeInstallType = NewInstall;
}
