#include "DapRegularRequestsController.h"

#include <QDebug>
#include <QRegularExpression>
#include <QJsonValue>
#include <QJsonArray>
#include <QJsonObject>
#include <QJsonDocument>

DapRegularRequestsController::DapRegularRequestsController(QObject *parent)
    : QObject(parent)
    , m_cmdList(new DapCommandList())
    , m_timerUpdateListWallets(new QTimer())
    , m_timerUpdateFee(new QTimer())
    , m_timerUpdateNetList(new QTimer())
{
    connect(this, &DapRegularRequestsController::needReinit, &DapNodePathManager::getInstance(), &DapNodePathManager::slotReinit);
}

DapRegularRequestsController::~DapRegularRequestsController()
{
    delete m_timerUpdateListWallets;
    delete m_timerUpdateFee;
    delete m_timerUpdateNetList;
    delete m_cmdList;
}

void DapRegularRequestsController::start()
{
    m_timerUpdateListWallets->start(WALLETLIST_COUNT);
    m_timerUpdateFee->start(FEE_TIMER);
    m_timerUpdateNetList->start(NETLIST_TIMER);

    connect(m_timerUpdateListWallets, &QTimer::timeout, this, &DapRegularRequestsController::updateListWallets);
    connect(m_timerUpdateFee, &QTimer::timeout, this, &DapRegularRequestsController::updateFeeInfo);
    connect(m_timerUpdateNetList, &QTimer::timeout, this, &DapRegularRequestsController::updateListNetworks);
    updateListNetworks();
    updateListWallets();
}

void DapRegularRequestsController::updateListNetworks()
{
    QStringList listNetworks;

    // Request
    auto respondVar = m_cmdList->getListNetworks();

    qDebug() << "[TEST] NetList: " << respondVar;

    // Parsing
    auto respond = m_cmdList->getJsonResult(respondVar,"net list");
    QString errorMsg = m_cmdList->getError(respond.error);
    if(!errorMsg.isEmpty())
    {
        qWarning() << "Get net list. Error:" << errorMsg;
        return;
    }

    auto networksObject = respond.result.toArray()[0].toObject();
    auto networksArray = networksObject["networks"].toArray();
    for(const auto& item: qAsConst(networksArray))
    {
        listNetworks.append(item.toString());
    }
    if(m_networkList != listNetworks)
    {
        m_networkList = listNetworks;
        updateFeeInfo();
    }
}

void DapRegularRequestsController::updateListWallets()
{
    QMap<QString, QString> listWallets;

    // Request
    auto respondVar = m_cmdList->getRequestListWallets();

    // Compare with last respond
    if(compareWithLastRespond(respondVar, m_lastRespondWallets)) return;

    // Parsing
    auto respond = m_cmdList->getJsonResult(respondVar,"wallet list");

    QString errorMsg = m_cmdList->getError(respond.error);
    if(!errorMsg.isEmpty())
    {
        qWarning() << "Get wallet list. Error:" << errorMsg;
        return;
    }

    auto wallets = respond.result[0].toArray();
    for(const auto& item: qAsConst(wallets))
    {
        auto walletObject = item.toObject();

        if(!walletObject.contains("Wallet"))
        {
            continue;
        }
        QString tmpName = walletObject["Wallet"].toString();
        auto tmpList = tmpName.split(".");
        if(tmpList.size() != 2)
        {
            continue;
        }
        if(tmpList[1] != "dwallet")
        {
            continue;
        }
        QString walletName = tmpList[0];
        if(!walletObject.contains("status"))
        {
            qWarning() << "The result is not a list.";
            continue;
        }
        QString status = walletObject["status"].toString();
        if(status == "protected-inactive")
        {
            status = "non-Active";
        }
        else if(status == "protected-active")
        {
            status = "Active";
        }
        else if(status == "unprotected")
        {
            status = "";
        }
        else
        {
            continue;
        }
        listWallets.insert(walletName, status);
    }

    m_listWallets = listWallets;
}

void DapRegularRequestsController::updateVersion()
{
    QString version = m_cmdList->getVersion();
    if(m_version != version)
    {
        emit needReinit();
        m_version = version;
        emit versionNodeUpadeted();
    }
}

bool DapRegularRequestsController::compareWithLastRespond(const QVariant &current, QByteArray &last)
{
    QByteArray curr = current.toJsonDocument().toJson();
    if(last == curr)
    {
        // is equal
        return true;
    }
    last = curr;
    return false;
}

void DapRegularRequestsController::updateFeeInfo()
{
    for(const auto& network: qAsConst(m_networkList))
    {
        auto request = m_cmdList->getJsonResult(m_cmdList->getFeeComand(network), "fee info").result;

        NetworkSpace::FeeInfoStruct result;

        auto testJsonBlock = [](bool expression, const QString& message) -> bool
        {
            if(expression)
            {
                qWarning() << QString("[DapRegularRequestsController] [getFee] The output of the node has changed or some kind of error has occurred.(%1)").arg(message);
                return false;
            }
            return true;
        };

        if(!testJsonBlock(request.isNull(), "the result block was not found"))
            continue;
        if(!testJsonBlock(!request.isArray(), "The block is no longer an array"))
            continue;

        auto mainArray = request.toArray();

        if(!testJsonBlock(mainArray.isEmpty(), "the array is empty"))
            continue;
        if(!testJsonBlock(!mainArray[0].toObject().contains("fees"), "the fees block is missing"))
            continue;

        auto fees = mainArray[0].toObject()["fees"].toObject();
        if(!testJsonBlock(!fees.contains("network"), "the network block is missing"))
            continue;

        {
            auto network = fees["network"].toObject();
            if(!testJsonBlock(!network.contains("addr"), "the network addr block is missing"))
                continue;
            if(!testJsonBlock(!network.contains("balance"), "the network balance block is missing"))
                continue;
            if(!testJsonBlock(!network.contains("coins"), "the network coins block is missing"))
                continue;
            if(!testJsonBlock(!network.contains("ticker"), "the network ticker block is missing"))
                continue;

            result.netCoins = network["coins"].toString();
            result.netDatoshi = network["balance"].toString();
            result.netTicker = network["ticker"].toString();
            result.netAddr = network["addr"].toString();
        }

        if(!testJsonBlock(!fees.contains("validators"), "the network block is missing"))
            continue;

        {
            auto validators = fees["validators"].toObject();

            auto getData = [&result, &testJsonBlock, &validators](const QString& blockName, bool nonNull = false)
            {
                auto object = validators[blockName].toObject();
                if(!testJsonBlock(!object.contains("balance"), "the validators balance block is missing"))
                    return;
                if(!testJsonBlock(!object.contains("coin"), "the validators coin block is missing"))
                    return;
                NetworkSpace::ValidatorItem item;
                QString coin = object["coin"].toString();
                QString datoshi = object["balance"].toString();
                if(nonNull)
                {
                    if(datoshi == "0")
                    {
                        item.coin = "0.05";
                        item.datoshi = "50000000000000000";
                        qWarning() << "[DapRegularRequestsController] fee is not valid. Block: " << blockName << " result: " << object;
                    }
                    else
                    {
                        item.coin = coin;
                        item.datoshi = datoshi;
                    }
                }
                else
                {
                    item.coin = coin;
                    item.datoshi = datoshi;
                }

                result.validatorFeeList.insert(blockName,item);
            };

            if(!testJsonBlock(!validators.contains("average"), "the validators average block is missing"))
                continue;
            getData("average", true);

            if(!testJsonBlock(!validators.contains("max"), "the validators balance block is missing"))
                continue;
            getData("max", true);

            if(!testJsonBlock(!validators.contains("median"), "the validators coins block is missing"))
                continue;
            getData("median", true);

            if(!testJsonBlock(!validators.contains("min"), "the validators ticker block is missing"))
                continue;
            getData("min");

            if(!testJsonBlock(!validators.contains("token"), "the validators addr block is missing"))
                continue;
            result.valTicker = validators["token"].toString();
        }
        m_feeInfoConteiner[network] = result;
    }
}

const QMap<QString, NetworkSpace::FeeInfoStruct> &DapRegularRequestsController::getFee() const
{
    return m_feeInfoConteiner;
}

const QString &DapRegularRequestsController::getVersion()
{
    if(m_version.simplified().isEmpty() )
        updateVersion();

    return m_version;
}

void DapRegularRequestsController::notifyWatcherRespond(QVariant nodeStatus)
{
    QString status;
    QJsonDocument jsonDoc = QJsonDocument::fromJson(nodeStatus.toByteArray());

    if (jsonDoc.isArray()) {
        QJsonArray jsonArray = jsonDoc.array();
        if (!jsonArray.isEmpty()) {
            QJsonValue jsonValue = jsonArray.first();
            if (jsonValue.isString()) {
                status = jsonValue.toString();
            }
        }
    }

    if(status == notifyStatusConnect)
        return;

    notifyStatusConnect = status;

    if(notifyStatusConnect == QAbstractSocket::SocketState::ConnectedState)
    {
        qDebug() << "[DapRegularRequestsController][notifyWatcherRespond] We are trying to get a list of networks.";
        updateListNetworks();
        updateVersion();
    }
    else
    {
        qDebug() << "[DapRegularRequestsController][notifyWatcherRespond] There is no way to get a list of networks.";
        m_networkList.clear();
        m_version.clear();
    }
}
