#include "TransactionQueue/DapTransactionQueueController.h"
#include <QVariant>
#include <QFile>
#include <QDateTime>
#include <QJsonObject>
//#include <QStringList>
//#include <dap_chain_common.h>
#include "../DapTypes/DapCoin.h"
#include "../DapTypes/DapExpConvert.h"

DapTransactionQueueController* DapTransactionQueueController::s_instance = nullptr;

DapTransactionQueueController::DapTransactionQueueController(QObject *parent)
    : QObject(parent)
    , m_requestTimer(new QTimer())
{
    loadFromFile();
}

DapTransactionQueueController::~DapTransactionQueueController()
{
    delete m_requestTimer;
}

DapTransactionQueueController* DapTransactionQueueController::getTransactionController()
{
    if(!s_instance)
    {
        s_instance = new DapTransactionQueueController();
    }
    return s_instance;
}

void DapTransactionQueueController::deleteTransactionController()
{
    if(s_instance)
    {
        delete s_instance;
        s_instance = nullptr;
    }
}

void DapTransactionQueueController::onInit()
{
    qDebug() << "[Queue] [DapTransactionQueueController] onInit";
    if(!m_queue.isEmpty())
    {
        tryStartTransactions();

    }
    m_requestTimer->start(TIMEOUT_REQUEST);
    connect(m_requestTimer, &QTimer::timeout, this, &DapTransactionQueueController::updateInfoHash);
    tryUpdateInfo();
}

void DapTransactionQueueController::tryStartTransactions()
{
    for(auto& network: m_queue.keys())
    {
        for(auto& wallet: m_queue[network].keys())
        {
            DataTransaction& item = m_queue[network][wallet].first();
            if(item.hash.isEmpty())
            {
                createTransaction(network, wallet);
            }
            else
            {
                emit newTransactionAdded(QStringList() << network << wallet << item.hash);
            }
        }
    }
}

void DapTransactionQueueController::updateInfoHash()
{
    for(auto& network: m_queue.keys())
    {
        for(auto& wallet: m_queue[network].keys())
        {
            DataTransaction& item = m_queue[network][wallet].first();
            if(!item.hash.isEmpty())
            {
                QString hash = item.hash;
                if(!hash.contains(REGULAR_HASH))
                {
                    qWarning() << QString("[Queue] [DapTransactionQueueController] An invalid hash was "
                                          "detected for a type %1 transaction for wallet %2 and network %3. "
                                          "The output may have changed at the command of this transaction. \ntx_hash: %4")
                                      .arg(m_queue[network][wallet][0].nameCommand).arg(wallet).arg(network).arg(hash);
                    clearCurrentTransaction(network, wallet);
                }
                else
                {
                    emit toGetTransactionData(QStringList()<< network << hash);
                }
            }
        }
    }
}

QJsonDocument DapTransactionQueueController::getJsonQueue()
{
    if(m_queue.isEmpty())
    {
        return {};
    }

    QJsonArray tmpMainArray;
    for (auto itemNetwork = m_queue.constBegin(); itemNetwork != m_queue.constEnd(); ++itemNetwork)
    {
        QJsonObject itemNetworkObject;
        QJsonArray tmpNetworkArray;
        for (auto itemWalletName = itemNetwork.value().constBegin(); itemWalletName != itemNetwork.value().constEnd(); ++itemWalletName)
        {
            QJsonObject itemObject;
            QJsonArray tmpArray;
            for(const DataTransaction& item: itemWalletName.value())
            {
                QJsonObject tmpObj;
                tmpObj.insert("queueHash", item.queueHash);
                tmpObj.insert("hash", item.hash);
                tmpObj.insert("nameCommand", item.nameCommand);
                tmpObj.insert("time", item.time);
                QJsonArray paramList;
                for(const QString& param: item.paramCommand)
                {
                    paramList.append(param);
                }
                tmpObj.insert("paramCommand", std::move(paramList));
                tmpArray.append(std::move(tmpObj));
            }
            itemObject.insert("key", itemWalletName.key());
            itemObject.insert("value", tmpArray);
            tmpNetworkArray.append(itemObject);
        }
        itemNetworkObject.insert("key", itemNetwork.key());
        itemNetworkObject.insert("value", tmpNetworkArray);
        tmpMainArray.append(itemNetworkObject);
    }

    return QJsonDocument(std::move(tmpMainArray));
}

void DapTransactionQueueController::queueFromJson(const QJsonDocument &document)
{
    if(document.isEmpty())
    {
        return;
    }
    m_queue.clear();
    QJsonArray mainArray = document.array();
    for(const QJsonValue& itemValue: qAsConst(mainArray))
    {
        QJsonObject itemObject = itemValue.toObject();
        if(!itemObject.contains("key") || !itemObject.contains("value"))
        {
            qWarning() << "[Queue] [DapTransactionQueueController] Has problem in  file struct";
            return;
        }

        QMap<QString, QVector<DataTransaction>> tmpQueue;
        QJsonArray conteinerWallets = itemObject["value"].toArray();
        for(const QJsonValue& walletsValue: qAsConst(conteinerWallets))
        {
            QJsonObject walletObject = walletsValue.toObject();
            if(!walletObject.contains("key") || !walletObject.contains("value"))
            {
                qWarning() << "[Queue] [DapTransactionQueueController] Has problem in  file struct";
                return;
            }
            QVector<DataTransaction> tmpValue;
            QJsonArray conteinerTransaction = walletObject["value"].toArray();
            for(const QJsonValue& transactionValue: conteinerTransaction)
            {
                QJsonObject transactionObject = transactionValue.toObject();
                if(!transactionObject.contains("hash") || !transactionObject.contains("nameCommand")
                    || !transactionObject.contains("time") || !transactionObject.contains("paramCommand"))
                {
                    qWarning() << "[Queue] [DapTransactionQueueController] Has a problrm in struct file";
                    return;
                }
                DataTransaction tmpTransaction;
                tmpTransaction.queueHash = transactionObject["queueHash"].toString();
                tmpTransaction.hash = transactionObject["hash"].toString();
                tmpTransaction.nameCommand = transactionObject["nameCommand"].toString();
                tmpTransaction.time = transactionObject["time"].toInt();
                for(const QJsonValue& paramValue: transactionObject["paramCommand"].toArray())
                {
                    tmpTransaction.paramCommand.append(paramValue.toString());
                }
                tmpValue.append(std::move(tmpTransaction));
            }
            tmpQueue.insert(walletObject["key"].toString(), std::move(tmpValue));
        }

        m_queue.insert(itemObject["key"].toString(), std::move(tmpQueue));
    }
}

void DapTransactionQueueController::saveToFile()
{
    auto jsonDoc = getJsonQueue();

    QFile file(FILE_NAME);
    if (!file.open(QIODevice::WriteOnly))
    {
        qDebug() << "[Queue] [DapTransactionQueueController] File is not created";
        return;
    }
    file.write(jsonDoc.toBinaryData());
    file.close();
}

void DapTransactionQueueController::loadFromFile()
{
    QFile file(FILE_NAME);
    if (!file.open(QIODevice::ReadOnly))
    {
        qDebug() << "File not opened";
        return ;
    }
    QByteArray jsonData = file.readAll();
    file.close();
    QJsonDocument jsonDocument = QJsonDocument::fromBinaryData(jsonData);
    if (jsonDocument.isNull())
    {
        qDebug() << "[Queue] [DapTransactionQueueController] Error when reading a JSON document from a file";
        return ;
    }
    else if (!jsonDocument.isArray())
    {
        qDebug() << "[Queue] [DapTransactionQueueController] A JSON document is not an array";
        return ;
    }
    queueFromJson(jsonDocument);
}

void DapTransactionQueueController::addTransactionHash(const QString& networkName, const QString& walletName, const QString& hash)
{
    m_mutex.lock();
    if(!m_queue.contains(networkName))
    {
        return;
    }
    if(!m_queue[networkName].contains(walletName))
    {
        return;
    }
    if(!hash.isEmpty())
    {
        emit newTransactionAdded(QStringList() << networkName << walletName << hash);
    }
    if(!hash.contains(REGULAR_HASH))
    {
        qWarning() << QString("[Queue] [DapTransactionQueueController] An invalid hash was "
                                                "detected for a type %1 transaction for wallet %2 and network %3. "
                                                "The output may have changed at the command of this transaction.")
                                            .arg(m_queue[networkName][walletName][0].nameCommand).arg(walletName).arg(networkName);
        clearCurrentTransaction(networkName, walletName);
    }
    else
    {
        m_queue[networkName][walletName][0].time = QDateTime::currentSecsSinceEpoch();
        m_queue[networkName][walletName][0].hash = hash;
        tryUpdateInfo();
        saveToFile();
    }
//    if(!m_requestTimer->isActive())
//    {
//        m_requestTimer->start(TIMEOUT_REQUEST);
//    }
    m_mutex.unlock();
}

bool DapTransactionQueueController::isReady(const QString& networkName, const QString& walletName)
{
    if(!m_queue.contains(networkName))
    {
        return true;
    }
    if(!m_queue[networkName].contains(walletName))
    {
        return true;
    }
    return m_queue[networkName][walletName].first().hash.isEmpty();
}

void DapTransactionQueueController::addNewTransaction(const QString& nameCommand, const QStringList& paramCommand,
                                                      const QString &networkName, const QString &walletName, const QString& hash
                                                      , const QString& queueHash)
{
    m_mutex.lock();

    if(!hash.isEmpty() && !hash.contains(REGULAR_HASH))
    {
        qWarning() << QString("[Queue] [DapTransactionQueueController] An invalid hash was "
                              "detected for a type %1 transaction for wallet %2 and network %3. "
                              "The output may have changed at the command of this transaction.")
                          .arg(nameCommand).arg(walletName).arg(networkName);
    }
    else
    {
        DataTransaction tmpData;
        tmpData.nameCommand = nameCommand;
        tmpData.paramCommand = paramCommand;
        tmpData.time = QDateTime::currentSecsSinceEpoch();
        tmpData.hash = hash;
        tmpData.queueHash = queueHash;

        if(!hash.isEmpty())
        {
            emit newTransactionAdded(QStringList() << networkName << walletName << hash);
        }

        if(m_queue.contains(networkName))
        {
            auto& netContainer = m_queue[networkName];
            if(netContainer.contains(walletName))
            {
                netContainer[walletName].append(std::move(tmpData));
            }
            else
            {
                QVector<DataTransaction> tmpVector = {std::move(tmpData)};
                netContainer.insert(walletName,std::move(tmpVector));
            }
        }
        else
        {
            QVector<DataTransaction> tmpVector = {std::move(tmpData)};
            QMap<QString, QVector<DataTransaction>> tmpMap = {{walletName, std::move(tmpVector)}};
            m_queue.insert(networkName, std::move(tmpMap));
        }
        //    if(!m_requestTimer->isActive())
        //    {
        //        m_requestTimer->start(TIMEOUT_REQUEST);
        //    }
        tryUpdateInfo();
        saveToFile();
    }

    m_mutex.unlock();
}

void DapTransactionQueueController::createTransaction(const QString& networkName, const QString &walletName)
{
    if(m_queue.isEmpty())
    {
        return;
    }
    if(m_queue.contains(networkName))
    {
        if(m_queue[networkName].contains(walletName))
        {
            DataTransaction& tmpData = m_queue[networkName][walletName].first();
            if(tmpData.hash.isEmpty() && !tmpData.isRequest)
            {
                tmpData.isRequest = true;
                qInfo() << QString("[Queue] [DapTransactionQueueController] Updating the wallet information. Network: %1, wallet name: %2").arg(networkName).arg(walletName);
                emit toGetWalletInfo(QStringList() << walletName << networkName);
            }
        }
    }
}

void DapTransactionQueueController::clearCurrentTransaction(const QString &networkName, const QString &walletName)
{
    qInfo() << QString("[Queue] [DapTransactionQueueController] The transaction on the %1 network will be deleted").arg(networkName);

    if(m_queue.contains(networkName))
    {
        if(m_queue[networkName].contains(walletName))
        {
            m_queue[networkName][walletName].removeFirst();
            if(m_queue[networkName][walletName].isEmpty())
            {
                m_queue[networkName].remove(walletName);
                if(m_queue[networkName].isEmpty())
                {
                    m_queue.remove(networkName);
                }
                saveToFile();
            }
            else
            {
                createTransaction(networkName, walletName);
            }
//            if(m_queue.isEmpty())
//            {
//                m_requestTimer->stop();
//            }
            tryUpdateInfo();
        }
        else
        {
            qWarning() << "[Queue] [DapTransactionQueueController] For some reason, no transactions were detected on this walletName";
        }
    }
    else
    {
        qWarning() << "[Queue] [DapTransactionQueueController] For some reason, no transactions were detected on this network";
    }
}

void DapTransactionQueueController::tryUpdateInfo()
{
    updateWalletsInfo();
    updateHistoryInfo();
}

void DapTransactionQueueController::updateHistoryInfo()
{
    QMap<QString,QJsonArray> result;
    if(m_queue.isEmpty())
    {
        emit updateHistoryForWallet(QVariant::fromValue(result));
        return;
    }

    auto appendTransaction = [&result](const QJsonObject& item, const QString& wallet)
    {
        if(!result.contains(wallet))
        {
            result.insert(wallet, QJsonArray());
        }
        result[wallet].append(item);
    };

    for(const auto& network: m_queue.keys())
    {
        for(const auto& wallet: m_queue[network].keys())
        {
            for(const DataTransaction& item: m_queue[network][wallet])
            {
                if(!item.hash.isEmpty())
                {
                    continue;
                }

                Q_ASSERT_X(m_dataParamsPool.contains(item.nameCommand), "DapTransactionQueueController", "[updateHistoryInfo] There is no definition of parameters.");

                auto& params = m_dataParamsPool[item.nameCommand];
                auto& paramCommand = item.paramCommand;
                QString walletName = params.walletNameCallback(paramCommand);
                QString address = params.addressCallback(paramCommand);
                QString token = params.tokenCallback(paramCommand);

                QString valueTransaction = params.valueCallback(paramCommand);
                if(DapExpConvert::isExp(valueTransaction))
                {
                    valueTransaction = DapExpConvert::expToFixed(valueTransaction);
                }
                QString value;
                if(valueTransaction.contains('.'))
                {
                    value = valueTransaction;
                }
                else
                {
                    uint256_t value256 = dap_uint256_scan_uninteger(valueTransaction.toStdString().data());
                    value = QString(dap_chain_balance_to_coins(value256));
                }
                uint256_t value_fee = dap_uint256_scan_uninteger(params.feeCallback(paramCommand).toStdString().data());
                QString valueFee = QString(dap_chain_balance_to_coins(value_fee));
                QString direction = params.directionCallback(paramCommand);

                QJsonObject itemObject;
                itemObject.insert("network", network);
                itemObject.insert("wallet_name", walletName);
                itemObject.insert("address", address);
                itemObject.insert("token", token);
                itemObject.insert("value", value);
                itemObject.insert("fee", valueFee);
                itemObject.insert("date_to_secs", QString::number(item.time));
                itemObject.insert("direction", direction);
                itemObject.insert("tx_status", "PROCESSING");
                QDateTime dateTime;
                dateTime.setSecsSinceEpoch(item.time);
                itemObject.insert("date", dateTime.toString("yyyy-MM-dd"));
                itemObject.insert("tx_hash", "UNKNOWN");
                itemObject.insert("status", "Queued");
                itemObject.insert("queueHash", item.queueHash);
                appendTransaction(std::move(itemObject), wallet);
            }
        }
    }

    emit updateHistoryForWallet(QVariant::fromValue(result));
}

void DapTransactionQueueController::transactionDataReceived(const QVariant& requestData)
{
    QJsonDocument replyDoc = QJsonDocument::fromJson(requestData.toByteArray());
    QJsonObject replyObj = replyDoc.object();
    QJsonObject result = replyObj["result"].toObject();

    if(!result.contains("hash") || !result.contains("net"))
    {
        qWarning() << "[Queue] [DapTransactionQueueController]Incorrect data.";
        return;
    }
    QString networkName = result["net"].toString();
    QString hash = result["hash"].toString();
    QString source;
    if(result.contains("source"))
    {
        source = result["source"].toString();
    }

    bool inMempool = source == "mempool";
    if(!inMempool)
    {
        qInfo() << QString("[Queue] The transaction with hash %1 from the %2 network is no longer in the mempool").arg(hash).arg(networkName);
    }
    else
    {
        qInfo() << QString("[Queue] Transaction with hash %1 from network %2 is still in mempool").arg(hash).arg(networkName);
    }  
    
    qint64 currentTime = QDateTime::currentMSecsSinceEpoch();
    if(m_queue.contains(networkName))
    {
        auto& item = m_queue[networkName];
        for(const auto& walletName: item.keys())
        {
            auto& itemFirst = item[walletName].first();
            if(itemFirst.hash != hash)
            {
                continue;
            }
            if(!inMempool || currentTime - itemFirst.time <= TIMEOUT_TRANSACTION)
            {
                if(isCurrentTransaction(networkName, walletName, hash))
                {
                    clearCurrentTransaction(networkName,walletName);
                }
            }
        }
    }
}

void DapTransactionQueueController::walletInfoReceived(const QVariant& requestData)
{
    QJsonDocument document = QJsonDocument::fromJson(requestData.toByteArray());
    QJsonObject object = document.object();
    if(!object.contains("result"))
    {
        qWarning() << "[Queue] [DapTransactionQueueController] Error there is no --result-- block in the data. " << object;
        return;
    }

    QJsonObject resultObject = object["result"].toObject();

    if(m_queue.isEmpty())
    {
        qDebug() << "[Queue] [DapTransactionQueueController] For some unknown reason, the queue is empty.";
        return;
    }

    if(!resultObject.contains("walletName"))
    {
        qWarning() << "[Queue] [DapTransactionQueueController] Error there is no -- walletName -- block in the data.";
        return;
    }

    if(!resultObject.contains("tokens"))
    {
        qWarning() << "[Queue] [DapTransactionQueueController] Error there is no -- tokens -- block in the data.";
        return;
    }

    if(!resultObject.contains("networkName"))
    {
        qWarning() << "[Queue] [DapTransactionQueueController] Error there is no -- networkName -- block in the data.";
        return;
    }

    QString networkName = resultObject["networkName"].toString();
    if(!m_queue.contains(networkName))
    {
        qDebug() << "[Queue] [DapTransactionQueueController] There is no queue for the network: " << networkName;
        return;
    }
    QString walletName = resultObject["walletName"].toString();
    if(!m_queue.contains(networkName))
    {
        return;
    }
    if(!m_queue[networkName].contains(walletName))
    {
        return;
    }
    auto& firstItem = m_queue[networkName][walletName].first();
    firstItem.isRequest = false;
    QString coins = "";

    Q_ASSERT_X(m_dataParamsPool.contains(firstItem.nameCommand), "DapTransactionQueueController", "[walletInfoReceived] There is no definition of parameters.");

    auto& params = m_dataParamsPool[firstItem.nameCommand];

    QString tokenNameTransaction = params.tokenCallback(firstItem.paramCommand);
    QString valueTransaction = params.valueCallback(firstItem.paramCommand);
    if(DapExpConvert::isExp(valueTransaction))
    {
        valueTransaction = DapExpConvert::expToFixed(valueTransaction);
    }
    if(DapExpConvert::isDatoshi(valueTransaction))
    {
        valueTransaction = DapExpConvert::datoshiToCoins(valueTransaction);
    }

    bool isTestExpenses = params.testExpenses();
    QString valFee = params.feeCallback(firstItem.paramCommand);
    if(DapExpConvert::isExp(valFee))
    {
        valFee = DapExpConvert::expToFixed(valFee);
    }
    if(DapExpConvert::isDatoshi(valFee))
    {
        valFee = DapExpConvert::datoshiToCoins(valFee);
    }

    auto& feeConteiner = params.feeInfo();
    bool isHasFee = false;
    NetworkSpace::FeeInfoStruct feeInfo;
    if(feeConteiner.contains(networkName))
    {
        feeInfo = feeConteiner[networkName];
        isHasFee = true;
    }
    else
    {
        qWarning() << "[Queue] [DapTransactionQueueController] No commissions were found for network " << networkName;
    }

    QJsonArray tokensArray = resultObject["tokens"].toArray();
    //   tiker    coin
    QMap<QString, QPair<QString, Dap::Coin>> tokenConteiner;

    for(const auto& itemValue: tokensArray)
    {
        QJsonObject item = itemValue.toObject();
        if(!item.contains("name"))
        {
            qWarning() << "[Queue] [DapTransactionQueueController] the token name was not found in the input block." << item;
            continue;
        }
        if(!item.contains("coins"))
        {
            qWarning() << "[Queue] [DapTransactionQueueController] the coins block was not found. " << item;
            return;
        }
        QString coin = item["coins"].toString();
        tokenConteiner.insert(item["name"].toString(), {coin, Dap::Coin(coin)});
    }

    Dap::Coin test = QString("0.0");

    auto checkBalance = [&tokenConteiner](const QString& tiker, const QString& valueCoin) ->bool
    {
        if(tokenConteiner.contains(tiker))
        {
            auto& coin = tokenConteiner[tiker].second;
            Dap::Coin value = valueCoin;
            if(coin > value)
            {
                coin = coin - value;
                return true;
            }
            else
            {
                qDebug() << "[Queue] [DapTransactionQueueController] wallet: " << tokenConteiner[tiker].first
                         << " Transaction value: " << valueCoin;
                qInfo() << "[Queue] [DapTransactionQueueController] Insufficient funds in the account.";
            }
        }
        else
        {
            qInfo() << "[Queue] [DapTransactionQueueController] Token 1 was not found in the wallet, the transaction will be deleted.";
        }
        return false;
    };

    if(isTestExpenses)
    {
        qDebug() << "[Queue] [DapTransactionQueueController] Check main value transaction.";
        if(!checkBalance(tokenNameTransaction, valueTransaction))
        {
            clearCurrentTransaction(networkName, walletName);
            return;
        }
    }
    QString netTicker, netCoin, valTicker, valCoin;
    if(isHasFee)
    {
        netTicker = feeInfo.netTicker;
        netCoin = feeInfo.netCoins;
        if(!netTicker.isEmpty() && !feeInfo.netCoins.isEmpty())
        {

            qDebug() << "[Queue] [DapTransactionQueueController] Check net fee.";
            if(!checkBalance(netTicker, netCoin))
            {
                clearCurrentTransaction(networkName, walletName);
                return;
            }
        }

        valTicker = feeInfo.valTicker;
        if(!feeInfo.validatorFeeList.isEmpty())
        {
            auto& valList = feeInfo.validatorFeeList;
            if(valList.contains("median"))
            {
                valCoin = valList["median"].coin;
            }
        }
    }

    if(!valTicker.isEmpty() && !valFee.isEmpty())
    {
        qDebug() << "[Queue] [DapTransactionQueueController] Check val fee.";
        if(!checkBalance(valTicker, valFee))
        {
            clearCurrentTransaction(networkName, walletName);
            return;
        }
    }

    qInfo() << QString("[Queue] [DapTransactionQueueController] %1 %2 will be charged.").arg(valueTransaction).arg(tokenNameTransaction);
    qInfo() << QString("[Queue] [DapTransactionQueueController] The validator's commission is recommended %1 %2 and is specified in the parameters %3 %2.")
                   .arg(valCoin).arg(valTicker).arg(valFee);
    qInfo() << QString("[Queue] [DapTransactionQueueController] The network commission is %1 %2.").arg(netCoin).arg(netTicker);

    qInfo() << "[Queue] [DapTransactionQueueController] There are enough funds in the account to create a transaction allowed.";
    emit toCreateTransaction(firstItem.nameCommand, firstItem.paramCommand);

}

void DapTransactionQueueController::updateInfoTransaction(const QVariant& requestData)
{
    auto params = requestData.toStringList();
    QString networkName = params[0];
    QString walletName = params[1];
    QString hash = params[2];
    qInfo() << QString("[Queue] [DapTransactionQueueController] There is a request to delete a transaction from the queue from the history side.");

    if(isCurrentTransaction(networkName, walletName, hash))
    {
        clearCurrentTransaction(networkName,walletName);
    }
}

bool DapTransactionQueueController::isCurrentTransaction(const QString& networkName, const QString &walletName, const QString &hash)
{
    if(m_queue.contains(networkName))
    {
        if(m_queue[networkName].contains(walletName))
        {
            if(m_queue[networkName][walletName].first().hash == hash && !m_queue[networkName][walletName].first().hash.isEmpty())
            {
                return true;
            }
        }
    }
    return false;
}

void DapTransactionQueueController::updateWalletsInfo()
{
    QVector<Command::QueueInfoConteiner> result;

    if(m_queue.isEmpty())
    {
        emit updateInfoForWallets(QVariant::fromValue(result));
        return;
    }

    for(auto& network: m_queue.keys())
    {
        for(auto& wallet: m_queue[network].keys())
        {
            for(auto& item: m_queue[network][wallet])
            {
                if(!item.hash.isEmpty())
                {
                    continue;
                }
                if(m_notChangeWalletInfoList.contains(item.nameCommand))
                {
                    continue;
                }
                Command::QueueInfoConteiner value;
                QStringList params = item.paramCommand;

                Q_ASSERT_X(m_dataParamsPool.contains(item.nameCommand), "DapTransactionQueueController", "[updateWalletsInfo] There is no definition of parameters.");

                auto& paramsCallback = m_dataParamsPool[item.nameCommand];
                QString valueTransaction = paramsCallback.valueCallback(params);
                if(DapExpConvert::isExp(valueTransaction))
                {
                    valueTransaction = DapExpConvert::expToFixed(valueTransaction);
                }
                value.value = valueTransaction;
                value.ticker = paramsCallback.tokenCallback(params);
                value.fee = paramsCallback.feeCallback(params);

                value.networkName = network;
                value.walletName = wallet;
                result.append(std::move(value));
            }
        }
    }

    emit updateInfoForWallets(QVariant::fromValue(result));
}

void DapTransactionQueueController::setParamsCommandCallback(const QString& commandName, ParamsFromCommand& callbacks)
{
    Q_ASSERT_X(callbacks.walletNameCallback && callbacks.addressCallback && callbacks.feeCallback && callbacks.tokenCallback && callbacks.valueCallback
               && callbacks.directionCallback && callbacks.testExpenses && callbacks.feeInfo,
               "DapTransactionQueueController", "[setParamsCommandCallback] There is no definition of parameters.");
    m_dataParamsPool.insert(commandName, callbacks);
}

void DapTransactionQueueController::addCommandWithoutInfoWallet(const QString& commandName)
{
    m_notChangeWalletInfoList.append(commandName);
}

QJsonObject DapTransactionQueueController::tryRemoveTransactionsWithParams(const QVariant &args)
{
    qDebug() << "[Queue] [DapTransactionQueueController] try remove transaction.";
    auto params = args.toStringList();
    QString error = "";
    auto getParams = [](const QStringList& list, const QString& paramName, QString& param) ->bool
    {
        if(list.contains(paramName))
        {
            int index = list.indexOf(paramName);
            if(list.size() < index + 2)
            {
                qWarning() << "A parameter may have been omitted: " << paramName;
                param.clear();
                return false;
            }
            if(!list[index+1].isEmpty())
            {
                param = list[index+1];
                return true;
            }
        }
        return false;
    };

    auto setError = [&](const QString& message)
    {
        qDebug() << "[Queue] [DapTransactionQueueController] " << message;
        error = message;
    };

    QString typeRemove, addParam;
    if(!getParams(params, "typeRemove", typeRemove) || typeRemove.isEmpty())
    {
        setError("The deletion scheme is not specified.");

    }

    getParams(params, "addParam", addParam);
    bool isHard = addParam == "hard";

    QJsonObject resultObject;

    if(typeRemove == "single")
    {
        QString idQueue = "";
        getParams(params, "idQueue", idQueue);
        if(!idQueue.isEmpty())
        {
            QString network = "";
            getParams(params, "network", network);
            QString wallet = "";
            getParams(params, "wallet", wallet);

            if(network.isEmpty() || wallet.isEmpty())
            {
                if(network.isEmpty()) setError("The network parameter is not specified.");
                if(wallet.isEmpty()) setError("The wallet parameter is not specified.");
            }
            else
            {
                if(m_queue.contains(network))
                {
                    auto& netQueue = m_queue[network];
                    if(netQueue.contains(wallet))
                    {
                        auto& walletQueue = netQueue[wallet];
                        int size = walletQueue.size();
                        for(int i = 0; i < size; i++)
                        {
                            if(walletQueue[i].queueHash != idQueue)
                            {
                                if(i == size - 1)
                                {
                                    setError("The transaction with id " + idQueue + " was not found.");
                                }
                                continue;
                            }

                            if(i == 0 && !isHard)
                            {
                                error = "The transaction got into the mempool and is now being tracked. If you still want to delete it, use the additional parameter addParam=hard.";
                                resultObject.insert("txHash", walletQueue[i].hash);
                                resultObject.insert("command", walletQueue[i].nameCommand);
                                break;
                            }

                            walletQueue.remove(i);
                            resultObject.insert("message", "All references to the transaction have been deleted.");
                            if(walletQueue.isEmpty())
                            {
                                m_queue[network].remove(wallet);
                            }
                            if(m_queue[network].isEmpty())
                            {
                                m_queue.remove(network);
                            }
                            saveToFile();
                            tryUpdateInfo();
                            if(i == 0 && !walletQueue.isEmpty())
                            {
                                createTransaction(network, wallet);
                            }
                            break;
                        }
                    }
                    else
                    {
                        setError("No queue found for wallet " + wallet);
                    }
                }
                else
                {
                    setError("No queue found for network " + network);
                }
            }
        }
        else
        {
            setError("The queue id is not specified.");
        }
    }
    else if(typeRemove == "all")
    {
        QString network = "";
        getParams(params, "network", network);
        QString wallet = "";
        getParams(params, "wallet", wallet);

        if(network.isEmpty() || wallet.isEmpty())
        {
            if(network.isEmpty()) setError("The network parameter is not specified.");
            if(wallet.isEmpty()) setError("The wallet parameter is not specified.");
        }
        else
        {
            if(m_queue.contains(network))
            {
                auto& netQueue = m_queue[network];
                if(netQueue.contains(wallet))
                {
                    auto& walletQueue = netQueue[wallet];
                    if(isHard)
                    {
                        walletQueue.clear();
                        netQueue.remove(wallet);
                        resultObject.insert("message", "The queue has been completely cleared, all transactions that remained in the mempool are now untraceable.");
                    }
                    else
                    {
                        int size = walletQueue.size();
                        if(size > 1)
                        {
                            walletQueue.remove(1, size - 1);
                        }
                        resultObject.insert("message", "All queues are cleared except for the first transactions that track transactions in the mempool.");
                    }
                    saveToFile();
                    tryUpdateInfo();
                }
                else
                {
                    setError("There is no queue for wallet " + wallet);
                }
            }
            else
            {
                setError("There is no queue for network " + network);
            }
        }
    }
    else
    {
        setError("There is no such type. type: " + typeRemove);
    }

    QJsonObject mainObj;
    if(!error.isEmpty())
    {
        mainObj.insert("error", error);
    }
    mainObj.insert("result", resultObject);
    return mainObj;
}

QVariant DapTransactionQueueController::tryRemoveTransactions(const QVariant &transactions)
{
    QList<QStringList> listTransaction;
    QVariantList lists = transactions.toList();
    for (const QVariant& listVariant : lists) {
        QStringList strList = listVariant.toStringList();
        listTransaction.append(strList);
    }

    int count = 0;

    for(auto& item: listTransaction)
    {
        if(item.size() != 3)
        {
            continue;
        }
        if(!m_queue.contains(item[0]))
        {
            continue;
        }
        auto& netTransactions = m_queue[item[0]];
        if(!netTransactions.contains(item[1]))
        {
            continue;
        }
        auto& walletTransaction = netTransactions[item[1]];
        bool isRemove = false;
        for(int i = 0; i < walletTransaction.size(); i++)
        {
            if(!walletTransaction[i].hash.isEmpty())
            {
                continue;
            }

            if(walletTransaction[i].time == item[2].toUInt())
            {
                walletTransaction.remove(i);
                count++;
                if(walletTransaction.isEmpty())
                {
                    netTransactions.remove(item[1]);
                }
                isRemove = true;
                break;
            }
        }
        if(isRemove)
        {
            if(netTransactions.isEmpty())
            {
                m_queue.remove(item[0]);
            }
            continue;
        }
    }
    QString result;
    if(listTransaction.size() == 1 && count == 1)
    {
        result = tr("The transaction has been deleted.");
    }
    else if(listTransaction.size() == count)
    {
        result = tr("Transactions have been deleted.");
    }
    else if(count == 0)
    {
        result = tr("Nothing has been deleted.");
    }
    else if(listTransaction.size() != count)
    {
        result = tr("Not all transactions have been deleted.");
    }

    saveToFile();
    tryUpdateInfo();

    QJsonDocument resultDoc(QJsonObject{{"result", result}});
    return resultDoc.toJson();
}

QVariant DapTransactionQueueController::isTtransactioninQueued(const QVariant& args)
{

    /// 0 - network
    /// 1 - wallet name
    /// 2 - public key

    QStringList argsList = args.toStringList();
    QString network = argsList[0];
    QString walletName = argsList[1];

    bool isQueued = false;

    if(m_queue.contains(network))
    {
        if(m_queue[network].contains(walletName))
        {
            const auto& queueList = m_queue[network][walletName];
            for(const DataTransaction& item: queueList)
            {
                if(item.hash.isEmpty() && item.paramCommand.contains("queueHash"))
                {
                    auto& params = item.paramCommand;
                    int index = params.indexOf("queueHash");
                    QString key = params[index + 1];
                    if(key == argsList[2])
                    {
                        isQueued = true;
                    }
                }
            }
        }
    }

    QJsonDocument resultDoc(QJsonObject{{"result", isQueued}});
    return resultDoc.toJson();
}
