#include "DapAbstractCommand.h"

/// Overloaded constructor.
/// @param asServiceName Service name.
/// @param parent Parent.
/// @details The parent must be either DapRPCSocket or DapRPCLocalServer.
/// @param asCliPath The path to cli nodes.
DapAbstractCommand::DapAbstractCommand(const QString &asServiceName, QObject *parent, const QString &asCliPath, const QString &asToolPath)
    : DapCommand(asServiceName, parent),
    m_parent(parent),
    cmdList(new DapCommandList(this))
{
    m_sCliPath =  &DapNodePathManager::getInstance().nodePaths.nodePath_cli;
    m_sToolPath = &DapNodePathManager::getInstance().nodePaths.nodePath_tool;

    if(PROG_TYPE != "GUI")
        qDebug()<<"INIT: " << asServiceName << ". CLI PATH = " << *m_sCliPath << "TOOL PATH = " << *m_sToolPath;

    connect(this, &DapAbstractCommand::toDataSignal, this, &DapAbstractCommand::replyToClient);
}

/// Send a notification to the client. At the same time, you should not expect a response from the client.
/// @details Performed on the service side.
/// @param arg1...arg10 Parameters.
void DapAbstractCommand::notifyToClient(const QVariant &args)
{
    //    QJsonDocument doc = QJsonDocument::fromVariant(args);

    DapRpcLocalServer * server = dynamic_cast<DapRpcLocalServer *>(m_parent);

    Q_ASSERT(server);

    DapRpcMessage request = DapRpcMessage::createNotification(QString("%1.%2").arg(this->getName()).arg("notifedFromService"), QJsonArray::fromVariantList(args.toList()));
    server->notifyConnectedClients(request);
}

/// Process the notification from the service on the client side.
/// @details Performed on the client side.
/// @param arg1...arg10 Parameters.
void DapAbstractCommand::notifedFromService(const QVariant &args)
{
    Q_UNUSED(args)

    emit clientNotifed(QVariant());
}

/// Send request to client.
/// @details Performed on the service side.
/// @param arg1...arg10 Parameters.
void DapAbstractCommand::requestToClient(const QVariant &args)
{
    QJsonDocument doc = QJsonDocument::fromVariant(args);

    DapRpcLocalServer * server = dynamic_cast<DapRpcLocalServer *>(m_parent);

    Q_ASSERT(server);

    DapRpcMessage request = DapRpcMessage::createRequest(QString("%1.%2").arg(this->getName()).arg("respondToService"), doc.toJson());
    DapRpcServiceReply * reply = server->notifyConnectedClients(request);
    connect(reply, SIGNAL(finished()), this, SLOT(replyFromClient()));
}

/// Send a response to the service.
/// @details Performed on the client side.
/// @param arg1...arg10 Parameters.
/// @return Reply to service.
QVariant DapAbstractCommand::respondToService(const QVariant &args)
{
    Q_UNUSED(args)

    return QVariant();
}

/// Reply from client.
/// @details Performed on the service side.
/// @return Client reply.
QVariant DapAbstractCommand::replyFromClient()
{
    DapRpcServiceReply *reply = static_cast<DapRpcServiceReply *>(sender());
    emit clientResponded(reply->response().toJsonValue().toVariant());
    return QVariant();
}

/// Send a notification to the service. At the same time, you should not expect a response from the service.
/// @details Performed on the client side.
/// @param arg1...arg10 Parameters.
void DapAbstractCommand::notifyToService(const QVariant &args)
{
    QJsonDocument doc = QJsonDocument::fromVariant(args);

    DapRpcSocket * socket = dynamic_cast<DapRpcSocket *>(m_parent);

    Q_ASSERT(socket);

    DapRpcMessage notify = DapRpcMessage::createNotification(QString("%1.%2").arg(this->getName()).arg("notifedFromClient"), doc.toJson());
    socket->notify(notify);
}

/// Process the notification from the client on the service side.
/// @details Performed on the service side.
/// @param arg1...arg10 Parameters.
void DapAbstractCommand::notifedFromClient(const QVariant &args)
{
    Q_UNUSED(args);

    emit serviceNotifed(QVariant());
}


Q_DECLARE_METATYPE(QVariantMap);

/// Send request to service.
/// @details Performed on the client side.
/// @param arg1...arg10 Parameters.
void DapAbstractCommand::requestToService(const QVariant &args)
{
    DapRpcServiceReply *reply = dynamic_cast<DapRpcSocket *>(m_parent)->invokeRemoteMethod(
        QString("%1.%2").arg(this->getName()).arg("respondToClient"), args);

    //qDebug() << "DapAbstractCommand::requestToService" << reply;
    connect(reply, SIGNAL(finished()), this, SLOT(replyFromService()));
}

/// Send a response to the client.
/// @details Performed on the service side.
/// @param arg1...arg10 Parameters.
/// @return Reply to client.
QVariant DapAbstractCommand::respondToClient(const QVariant &args)
{
    QStringList listArgs;
    for(const QString &str: args.toStringList())
    {
        if(!str.trimmed().isEmpty())
            listArgs.append(str);
    }

    //Q_UNUSED(args)
//    if(listArgs.count() >= 3)
    if(this->getName() == "DapAddWalletCommand")
    {
        if(listArgs.count() > 3)
            listArgs[3] = "****"; //hide pass in logs
    }

    qDebug() << "[" << this->getName() << "] [respondToClient] " << listArgs;
    return QVariant();
}

/// Reply from service.
/// @details Performed on the service side.
/// @return Service reply.
void DapAbstractCommand::replyFromService()
{
    DapRpcServiceReply *reply = static_cast<DapRpcServiceReply *>(sender());

    emit serviceResponded(reply->response().toJsonValue().toVariant());
}

QString DapAbstractCommand::requestToNode(QString cmd){

    QProcess process;
    process.start(cmd);

    if(process.waitForFinished(10000))
        return QString::fromLatin1(process.readAll());
    else{
        qWarning()<<QString("Timeout request: %1").arg(cmd);
        return QString();
    }
}

void DapAbstractCommand::replyToClient(const QVariant &args)
{
    QVariant result = respondToClient(args);

    emit dataGetedSignal(std::move(result));
}

void DapAbstractCommand::addWeb3Result(QJsonObject& result, const QVariant &args)
{
    QStringList list = args.toStringList();
    if(!list.contains(WEB3_KEY))
    {
        return;
    }
    int index = list.indexOf(WEB3_KEY);
    result.insert(WEB3_KEY, list[index + 1]);
}

QString DapAbstractCommand::getParamsByKey(const QString& key, const  QStringList& params)
{
    if(!params.contains(key))
    {
        return QString();
    }
    int index = params.indexOf(key);
    int size = params.size();
    if(index + 1 >=  size)
    {
        qWarning() << "[DapAbstractCommand] There are not enough parameters";
        return QString();
    }
    return params[index + 1];
}

QMap<QString, NetworkSpace::FeeInfoStruct> DapAbstractCommand::getFee() const
{
    if(m_regularController)
    {
        return m_regularController->getFee();
    }
    return QMap<QString, NetworkSpace::FeeInfoStruct>();
}

void DapAbstractCommand::setRegularController(DapRegularRequestsController* controller)
{
    m_regularController = controller;
    initRegularRequestController();
}

QMap<QString, QString> DapAbstractCommand::getListWallets() const
{
    if(m_regularController)
    {
        return m_regularController->getListWallets();
    }
    return QMap<QString, QString>();
}

const QStringList& DapAbstractCommand::getListNetworks() const
{
    if(m_regularController)
    {
        return m_regularController->getNetworkList();
    }
    return {};
}

const QString& DapAbstractCommand::getVersion() const
{
    if(m_regularController)
    {
        return m_regularController->getVersion();
    }
    return QString();
}
