/****************************************************************************
**
** This file is part of the libdap-qt-ui-chain-wallet library.
**
** The class implements the command interface.
**
****************************************************************************/

#ifndef DAPABSTRACTCOMMAND_H
#define DAPABSTRACTCOMMAND_H

#include <QObject>
#include <QVariant>
#include <QProcess>
#include <QRegularExpression>

#include "DapRpcSocket.h"
#include "DapRpcLocalServer.h"
#include "DapCommandList.h"
#include "RequestController/DapRegularRequestsController.h"
#include "DapNodePathManager.h"


typedef  DapRpcService DapCommand;

class DapAbstractCommand : public DapCommand
{
    Q_OBJECT

public:
    void setRegularController(DapRegularRequestsController *controller);

signals:
    /// The signal is emitted in case of successful notification of the client by the service.
    /// @param aNotify Notification.
    void clientNotifed(const QVariant& aNotify);
    /// The signal is emitted if the client has successfully received
    /// a request from the service and responded to the service.
    /// @param asRespond Client response.
    void clientResponded(const QVariant& aRespond);
    /// The signal is emitted if the client successfully notifies the service.
    /// @param aNotify Notification.
    void serviceNotifed(const QVariant& aNotify);
    /// The signal is emitted if the service has successfully received
    /// a request from the client and responded to the client
    /// @param asRespond Service response.
    void serviceResponded(const QVariant& aRespond);

    void nodeReply(const QString& reply);

    /// This signal is for the case if we are working without a service
    /// and just need to get a response from the node
    void toDataSignal(const QVariant &args);
    /// A signal sending a response from the node and converted in the right way
    void dataGetedSignal(const QVariant& reply);

public slots:
    /// Send a notification to the client. At the same time, you should not expect a response from the client.
    /// @details Performed on the service side.
    /// @param arg1...arg10 Parameters.
    Q_INVOKABLE virtual void notifyToClient(const QVariant &args = QVariant());
    /// Process the notification from the service on the client side.
    /// @details Performed on the client side.
    /// @param arg1...arg10 Parameters.
    Q_INVOKABLE virtual void notifedFromService(const QVariant &args = QVariant());
    /// Send request to client.
    /// @details Performed on the service side.
    /// @param arg1...arg10 Parameters.
    Q_INVOKABLE virtual void requestToClient(const QVariant &args = QVariant());
    /// Send a response to the service.
    /// @details Performed on the client side.
    /// @param arg1...arg10 Parameters.
    /// @return Reply to service.
    virtual QVariant respondToService(const QVariant &args = QVariant());
    /// Reply from client.
    /// @details Performed on the service side.
    /// @return Client reply.
    virtual QVariant replyFromClient();

    /// The slot is needed for multithreading so that all 
    /// processing takes place in the service stream. 
    void replyToClient(const QVariant &args);

    /// Send a notification to the service. At the same time, you should not expect a response from the service.
    /// @details Performed on the client side.
    /// @param arg1...arg10 Parameters.
    Q_INVOKABLE virtual void notifyToService(const QVariant &args = QVariant());
    /// Process the notification from the client on the service side.
    /// @details Performed on the service side.
    /// @param arg1...arg10 Parameters.
    Q_INVOKABLE virtual void notifedFromClient(const QVariant &args = QVariant());
    /// Send request to service.
    /// @details Performed on the client side.
    /// @param arg1...arg10 Parameters.
    Q_INVOKABLE virtual void requestToService(const QVariant &args = QVariant());
    /// Send a response to the client.
    /// @details Performed on the service side.
    /// @param arg1...arg10 Parameters.
    /// @return Reply to client.
    virtual QVariant respondToClient(const QVariant &args = QVariant());
    /// Reply from service.
    /// @details Performed on the service side.
    /// @return Service reply.
    virtual void replyFromService();

    QString requestToNode(QString cmd);
protected:
    /// Overloaded constructor.
    /// @param asServiceName Service name.
    /// @param parent Parent.
    /// @details The parent must be either DapRPCSocket or DapRPCLocalServer.
    /// @param asCliPath The path to cli nodes.
    explicit DapAbstractCommand(const QString &asServiceName, QObject *parent = nullptr, const QString &asCliPath = QString(), const QString &asToolPath = QString());

    void addWeb3Result(QJsonObject& result, const QVariant& args);

    virtual void initRegularRequestController() {}

    const QString& getVersion() const;
    const QStringList& getListNetworks() const;
    QMap<QString,QString> getListWallets() const;
    QMap<QString, NetworkSpace::FeeInfoStruct> getFee() const;
    QString getParamsByKey(const QString& key, const  QStringList& params);
protected:
    /// Parent.
    QObject * m_parent {nullptr};
    /// The path to cli nodes.
    QString *m_sCliPath;
    QString *m_sToolPath;
    /// Common result key for json-rpc
    const QString RESULT_KEY = "result";

    const QString WEB3_KEY = "web3";

    const QString ERROR_KEY = "error";
    const QString WALLET_NON_ACTIVE_KEY = "non-Active";

    /// Ptr to commandList class
    DapCommandList* cmdList;

    DapRegularRequestsController* m_regularController = nullptr;
};

#endif // DAPABSTRACTCOMMAND_H

