#include "DapCommandList.h"
#include <QRegExp>
#include <QRegularExpression>
#include "dapconfigreader.h"
#include <cmath>

#include "DapNodePathManager.h"

DapCommandList::DapCommandList(QObject *parent)
    : QObject(parent)
{
    initPaths();
    connect(&DapNodePathManager::getInstance(), &DapNodePathManager::signalReinitFinish, this, &DapCommandList::initPaths);
}

void DapCommandList::initPaths()
{
    m_cliPath =  &DapNodePathManager::getInstance().nodePaths.nodePath_cli;
    m_toolPath = &DapNodePathManager::getInstance().nodePaths.nodePath_tool;
    m_nodePath = &DapNodePathManager::getInstance().nodePaths.nodePath;
    // m_cfgToolCtrl;

    updateConfigParams();
}

/// List all networks or list all chains in selected network
QVariant DapCommandList::getListNetworks()
{
    auto command = QString("net list");

    return requestToNode(QPair<QString*, QString>(m_cliPath, command), Dap::RequestType::JSON);
}

/// Create new wallet
QVariant DapCommandList::addWallet(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("wallet new -w %1 -sign %2")
                       .arg(params[0]) // wallet name
                       .arg(params[1]); // sign

    if (!params[2].isEmpty())
        command += " -restore "+params[2];

    if(!params[3].isEmpty())
        command += " -password "+params[3];

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

/// Create password for wallet
QVariant DapCommandList::createPassForWallet(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("wallet convert -w %1 -password %2")
                       .arg(params[0]) // wallet name
                       .arg(params[1]); // password


    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

/// Displays the current fee
QVariant DapCommandList::getFeeComand(const QString &net)
{
    auto command = QString("net -net %1 get fee").arg(net);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

NetworkSpace::FeeInfoStruct DapCommandList::getFee(const QString &net)
{
    auto request = getJsonResult(getFeeComand(net), "ledger info");
    NetworkSpace::FeeInfoStruct result;

    auto testJsonBlock = [](bool expression, const QString& message) -> bool
    {
        if(expression)
        {
            qWarning() << QString("[getFee] The output of the node has changed or some kind of error has occurred.(%1)").arg(message);
            return false;
        }
        return true;
    };

    if(!testJsonBlock(request.result.isNull(), "the result block was not found"))
        return result;
    if(!testJsonBlock(!request.result.isArray(), "The block is no longer an array"))
        return result;

    auto mainArray = request.result.toArray();

    if(!testJsonBlock(mainArray.isEmpty(), "the array is empty"))
        return result;
    if(!testJsonBlock(!mainArray[0].toObject().contains("fees"), "the fees block is missing"))
        return result;

    auto fees = mainArray[0].toObject()["fees"].toObject();
    if(!testJsonBlock(!fees.contains("network"), "the network block is missing"))
        return result;

    {
        auto network = fees["network"].toObject();
        if(!testJsonBlock(!network.contains("addr"), "the network addr block is missing"))
            return result;
        if(!testJsonBlock(!network.contains("balance"), "the network balance block is missing"))
            return result;
        if(!testJsonBlock(!network.contains("coins"), "the network coins block is missing"))
            return result;
        if(!testJsonBlock(!network.contains("ticker"), "the network ticker block is missing"))
            return result;

        result.netCoins = network["coins"].toString();
        result.netDatoshi = network["balance"].toString();
        result.netTicker = network["ticker"].toString();
        result.netAddr = network["addr"].toString();
    }

    if(!testJsonBlock(!fees.contains("validators"), "the network block is missing"))
        return result;

    {
        auto validators = fees["validators"].toObject();

        auto getData = [&result, &testJsonBlock, &validators](const QString& blockName)
        {
            auto object = validators[blockName].toObject();
            if(!testJsonBlock(!object.contains("balance"), "the validators balance block is missing"))
                return;
            if(!testJsonBlock(!object.contains("coin"), "the validators coin block is missing"))
                return;
            NetworkSpace::ValidatorItem item;
            item.coin = object["coin"].toString();
            item.datoshi = object["balance"].toString();
            result.validatorFeeList.insert(blockName,item);
        };

        if(!testJsonBlock(!validators.contains("average"), "the validators average block is missing"))
            return result;
        getData("average");

        if(!testJsonBlock(!validators.contains("max"), "the validators balance block is missing"))
            return result;
        getData("max");

        if(!testJsonBlock(!validators.contains("median"), "the validators coins block is missing"))
            return result;
        getData("median");

        if(!testJsonBlock(!validators.contains("min"), "the validators ticker block is missing"))
            return result;
        getData("min");

        if(!testJsonBlock(!validators.contains("token"), "the validators addr block is missing"))
            return result;
        result.valTicker = validators["token"].toString();
    }

    return result;
}

/// Orders list
QVariant DapCommandList::getListOrders(const QVariant &args, QString &error)
{
    QStringList params = args.toStringList();
    QString tmpParam;
    QString command;

    if(getParamForPairList(params, "net", tmpParam))
    {
        command = QString("net_srv -net %1 order find").arg(tmpParam);
    }
    else
    {
        error = "Error. Network argument is empty";
        return QString();
    }

    if(getParamForPairList(params, "direction", tmpParam)) command += QString(" -direction %1").arg(tmpParam);
    if(getParamForPairList(params, "srv_uid",   tmpParam)) command += QString(" -srv_uid %1").arg(tmpParam);
    if(getParamForPairList(params, "unit",      tmpParam)) command += QString(" -price_unit %1").arg(tmpParam);
    if(getParamForPairList(params, "tokenName", tmpParam)) command += QString(" -price_token %1").arg(tmpParam);
    if(getParamForPairList(params, "price_min", tmpParam)) command += QString(" -price_min %1").arg(tmpParam);
    if(getParamForPairList(params, "price_max", tmpParam)) command += QString(" -price_max %1").arg(tmpParam);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::getServiceLimits(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("net_srv get_limits -net %1 -srv_uid %2 -provider_pkey_hash %3 -client_pkey_hash %4")
                       .arg(params[0])  // net
                       .arg(params[1])  // srv_uid
                       .arg(params[2])  // pr_hash
                       .arg(params[3]); // cl_hash

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

/// Ledger information
QVariant DapCommandList::getListTokensCommand(const QString &net)
{
    auto command = QString("ledger list coins -net %1").arg(net);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

QMap<QString, LadgerSpace::TokensData> DapCommandList::getListTokens(const QString &net)
{
    QString respond = getListTokensCommand(net).toString();

    QMap<QString, LadgerSpace::TokensData> result;

    auto tokensInfo = respond.split("-->");

    for(auto& tokenStr: tokensInfo)
    {
        auto tokenDetailInfoList = tokenStr.split('\n');
        LadgerSpace::TokensData tokenData;
        for( int i = 0; i < tokenDetailInfoList.size(); i++)
        {
            if(tokenDetailInfoList[i].contains("line"))
            {
                LadgerSpace::StatusToken status;
                i++;
                Dap::getValue<LadgerSpace::StatusToken>(status, tokenDetailInfoList[i]);
                i++;
                Dap::getValue<LadgerSpace::StatusToken>(status, tokenDetailInfoList[i]);
                i++;
                Dap::getValue<LadgerSpace::StatusToken>(status, tokenDetailInfoList[i]);
                if(status.hash.isEmpty() || status.signType.isEmpty() || status.bytes.isEmpty())
                {
                    qWarning() << "there are problems with status parsing";
                }
                tokenData.statusList.append(std::move(status));
            }
            Dap::getValue<LadgerSpace::TokensData>(tokenData, tokenDetailInfoList[i]);
        }
        if(!tokenData.tokenName.isEmpty())
        {
            result.insert(tokenData.tokenName,tokenData);
        }
    }

    return result;
}

/// Ledger tx hash
QVariant DapCommandList::getLedgerTxHashCommand(const QString &net, const QString &hash, QString &error)
{
    auto command = QString("ledger info -hash %1 -net %2")
                       .arg(hash).arg(net);

    auto request = getJsonResult(requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON), "ledger info");

    QJsonObject resultObj;

    if(request.result.isNull())
    {
        error = "Error";
    }
    else
    {
        resultObj = request.result.toArray()[0].toObject();
    }

    QJsonDocument resultDoc(resultObj);
    return resultDoc.toVariant();
}

/// Get request list of wallets
QVariant DapCommandList::getRequestListWallets() {

    QString command = QString("wallet list");

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

/// Make transaction
QVariant DapCommandList::createTransaction(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("tx_create -net %1 -from_wallet %2 -to_addr %3 -token %4 -value %5 -fee %6 ")
                       .arg(params[0]) // net
                       .arg(params[1]) // wallet name
                       .arg(params[2]) // address
                       .arg(params[3]) // token
                       .arg(params[4]) // value
                       .arg(params[5]); // commission
    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::srvStakeDelegate(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("srv_stake delegate -cert %1 -net %2 -w %3 -value %4 -fee %5")
                       .arg(params[0]) // cert name
                       .arg(params[1]) // network
                       .arg(params[2]) // wallet_name
                       .arg(params[3]) // value
                       .arg(params[4]); // fee

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::srvStakeInvalidate(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("srv_stake invalidate -net %1 -tx %2 -w %3 -fee %4")
                       .arg(params[0]) // network
                       .arg(params[1]) // hash
                       .arg(params[2]) // wallet_name
                       .arg(params[3]); // fee

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::createJsonTransaction(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("tx_create_json -json %1")
                       .arg(params[0]); // path

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

/// Displays the current net status
QVariant DapCommandList::getNetworkStateCommand(const QString &net)
{
    if(net.isEmpty())
    {
        qWarning() << "getNetworkState() net name is empty";
        return QString();
    }
    auto command = QString("net -net %1 get status").arg(net);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

/// Displays the current net status
QVariant DapCommandList::getNetworkIdCommand(const QString &net)
{
    if(net.isEmpty())
    {
        qWarning() << "getNetworkId() net name is empty";
        return QString();
    }
    auto command = QString("net -net %1 get id").arg(net);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

NetworkSpace::NetworkStatus DapCommandList::getNetworkState(const QString &net, QString &error)
{
    NetworkSpace::NetworkStatus result;

    if(net.isEmpty())
    {
        error = "Argument of network is empty";
        return result;
    }

    auto respondCommand = getNetworkStateCommand(net);
    qDebug() << "[TEST] net: " << net;
    qDebug() << "[TEST] respondCommand: " << respondCommand;
    auto respond = getJsonResult(respondCommand, "net get status");

    if(respond.result.isNull())
    {
        error = "Error." + getError(respond.error);
        return result;
    }

    auto resultObj = respond.result.toArray()[0].toObject();

    if(!resultObj.contains("status"))
    {
        error = "Error." + getError(respond.result);
        return result;
    }

    auto checkContains = [](const QJsonObject& obj, const QString& key) -> bool
    {
        if(!obj.contains(key))
        {
            qWarning() << QString("There are problems in the output. The parameter with the key '%1' was not found in the response.").arg(key);
            return false;
        }
        else
        {
            return true;
        }
    };

    auto getChainPercent = [](const QJsonObject& obj, const QString& chain) -> double
    {
        double chainPercent = 0.0;
        QJsonObject chainObj = obj[chain].toObject();
        if(chainObj.contains("status"))
        {
            QString status = chainObj["status"].toString();
            if(status == "synced")
            {
                chainPercent = 100.0;
            }
            else if(status == "not synced")
            {
                chainPercent = 0.0;
            }
            else
            {
                if(chainObj.contains("percent"))
                {
                    bool ok;
                    QString percntStr = chainObj["percent"].toString();
                    percntStr = percntStr.remove("%").trimmed();
                    chainPercent = percntStr.toDouble(&ok);
                    if(!ok)
                    {
                        qDebug() << "Value of 'percent' is not a number:" << percntStr;
                    }
                }
                else
                {
                    qDebug() << "No have key 'percent' in the chain:" << chain;
                }
            }
        }
        else
        {
            qDebug() << "No have key of chain with name:" << chain;
        }
        return chainPercent;
    };

    auto statusObj = resultObj["status"].toObject();

    if(checkContains(statusObj, "net"))
    {
        result.network = statusObj["net"].toString();
    }
    if(checkContains(statusObj, "current_addr"))
    {
        result.currentAddress = statusObj["current_addr"].toString();
    }
    if(checkContains(statusObj, "links"))
    {
        auto linksObj = statusObj["links"].toObject();
        if(checkContains(linksObj, "active")) result.activeLinks =  QString::number(linksObj["active"].toInt());
        if(checkContains(linksObj, "required")) result.totalLinks = QString::number(linksObj["required"].toInt());
    }
    if(checkContains(statusObj, "states"))
    {
        auto statesObj = statusObj["states"].toObject();
        if(checkContains(statesObj, "current")) result.currantState = statesObj["current"].toString();
        if(checkContains(statesObj, "target")) result.targetState = statesObj["target"].toString();
    }
    if(checkContains(statusObj, "processed"))
    {
        double summPercent = 0.0;

        auto syncObj = statusObj["processed"].toObject();
        if(checkContains(syncObj, "main") && checkContains(syncObj, "zerochain"))
        {
            double mainPercent = getChainPercent(syncObj, "main");
            double zeroPercent = getChainPercent(syncObj, "zerochain");
            summPercent = mainPercent * 0.45 + zeroPercent * 0.45;
        }
        else
        {
            qDebug() << "No have keys with name of chains";
        }

        result.syncPercent = QString::number(static_cast<int>(std::round(summPercent)));
    }

    if(result.targetState.isEmpty() ||
        result.currantState.isEmpty() ||
        result.currentAddress.isEmpty() ||
        result.network.isEmpty())
    {
        error = "There are problems in the output";
        qWarning() << error;
    }

    return result;
}

/// Display information of wallet
QVariant DapCommandList::getWalletInfoCommand(const QString &net, const QString &wallet, QString &error)
{
    if(wallet.isEmpty())
    {
        error = "getWalletInfo() wallet is empty";
        return {};
    }
    else if(net.isEmpty())
    {
        error = "getWalletInfo() net name is empty";
        return {};
    }

    auto command = QString("wallet info -w %1 -net %2").arg(wallet).arg(net);

    auto resultJson = getJsonResult(requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON), "wallet info");

    auto result = resultJson.result[0].toArray()[0].toObject();

    QJsonDocument resultDoc(result);
    return resultDoc.toVariant();
}

DapWalletSpace::WalletInfoType DapCommandList::getWalletInfo(const QString &net, const QString &wallet, QString *error)
{
    DapWalletSpace::WalletInfoType result;
    DapWalletSpace::BalanceType token;
    QString errorMsg;

    auto request = getWalletInfoCommand(net, wallet, errorMsg).toJsonObject();

    if(!errorMsg.isEmpty())
    {
        qWarning() << errorMsg;
        if(error) *error = errorMsg;
        return result;
    }

    result.walletName = request["nwallet"].toString();
    result.address = request["addr"].toString();
    result.networkName = request["network"].toString();

    auto tokensArray = request["tokens"].toArray();
    for(auto item: tokensArray)
    {
        QJsonObject obj = item.toObject();
        token.coins = obj["coins"].toString();
        token.datoshi = obj["datoshi"].toString();
        auto tokenObj = obj["token"].toObject();
        if(tokenObj.contains("ticker"))
        {
            token.token = tokenObj["ticker"].toString();
        }
        else
        {
            qWarning() << "The ticker was not found";
        }
        if(tokenObj.contains("description"))
        {
            token.description = tokenObj["description"].toString();
        }
        else
        {
            qWarning() << "The description was not found";
        }

        if(!token.token.isEmpty() && !token.coins.isEmpty() && !token.datoshi.isEmpty())
        {
            result.tokenList.append(token);
            token.token.clear();
            token.coins.clear();
            token.datoshi.clear();
        }
        else
        {
            qWarning() << "There are problems when parsing tokens";
        }
    }
    return result;
}

/// Transaction history by wallet name
QVariant DapCommandList::getTXHistory(const QString &net, const QString &walletName, const QString &walletAddr)
{
    bool byName = walletName.isEmpty() ? false : true;
    QString wallet = byName ? walletName : walletAddr;
    if(wallet.isEmpty())
    {
        qWarning() << "getTXHistory() wallet is empty";
        return QString();
    }
    else if(net.isEmpty())
    {
        qWarning() << "getTXHistory() net name is empty";
        return QString();
    }
    auto command = QString("tx_history -net %1 %2 %3").arg(net).arg(byName ? "-w" : "-addr").arg(wallet);

    QVariant result = requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);

    return result;
}

/// Get the exchange orders list within specified net name
QVariant DapCommandList::getXchangeOrdersList(const QString &net)
{
    if(net.isNull() || net.isEmpty())
    {
        qWarning() << "getXchangeOrdersList() net name is empty";
        return QString();
    }
    auto command = QString("srv_xchange orders -net %1").arg(net);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

/// List of all token pairs
QVariant DapCommandList::getXchangeTokenPairsList(const QString &net)
{
    if(net.isEmpty())
    {
        qWarning() << "getXchangeTokenPairsList() net name is empty";
        return QString();
    }
    auto command = QString("srv_xchange token_pair -net %1 list all")
                       .arg(net);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

/// Get average rate for token pair
QPair<qint64, QString> DapCommandList::getXchangeAverage(const QVariant &args)
{
    auto nodeAverage = getXchangeAverageCommand(args);

    auto resultStruct = getJsonResult(nodeAverage, "getXchangeAverage current rate");

    QPair<qint64, QString> resultPair{0, "0.0"} ;
    auto result = resultStruct.result;
    if(result.isNull())
    {
        // qWarning() << "";
        return resultPair;
    }
    QJsonArray level1 = result.toArray();

    if(level1.isEmpty())
    {
        return resultPair;
    }

    QJsonObject pairObject = level1[0].toObject();
    if(pairObject.isEmpty())
    {
        return resultPair;
    }

    if(pairObject.contains("errors"))
    {
        return resultPair;
    }
    if(!pairObject.contains("Last rate") || !pairObject.contains("Last rate time"))
    {
        qWarning() << "[getXchangeAverage] Either the structure or the names of the keys have changed.";
        return resultPair;
    }
    QDateTime date = QDateTime::fromString(pairObject["Last rate time"].toString(), Qt::RFC2822Date).toLocalTime();
    date.setTimeSpec(Qt::LocalTime);
    qint64 dateToSecs = date.toSecsSinceEpoch();
    resultPair.first = dateToSecs;
    resultPair.second = pairObject["Last rate"].toString();
    return resultPair;
}

QVariant DapCommandList::getXchangeAverageCommand(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("srv_xchange token_pair -net %1 rate average"
                           " -token_from %3 -token_to %2")
                       .arg(params[0]) // net
                       .arg(params[1]) // token to
                       .arg(params[2]); // token from

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

/// Print rate history for token pair
QVariant DapCommandList::getXchangeHistory(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("srv_xchange token_pair -net %1 rate history"
                           " -token_from %3 -token_to %2")
                       .arg(params[0]) // net
                       .arg(params[1]) // token from
                       .arg(params[2]); // token to
    if(params.size() > 3)
    {
        if(!(params[3].isEmpty() && params[4].isEmpty()))
            command += QString(" -time_from %1 -time_to %2")
                           .arg(params[3]) // time from
                           .arg(params[4]); // time to
    }


    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

/// List of exchange transactions
QVariant DapCommandList::getXchangeList(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("srv_xchange tx_list -net %1")
                       .arg(params[1]); // net

    if(params[0] == "GetOrdersPrivate")
        command += QString(" -addr %1 ")
                       .arg(params[2]); // address
    if(params.size() > 3)
    {
        if(!(params[3].isEmpty() && params[4].isEmpty()))
            command += QString(" -time_from %1 -time_to %2")
                           .arg(params[3]) // time from
                           .arg(params[4]); // time to
    }
    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

/// Proc mempool entrie with specified hash for selected chain network
QVariant DapCommandList::mempoolProcess(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("mempool_proc -net %1 -chain %2")
                       .arg(params[0]) // net
                       .arg(params[1]); // chain

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

/// Find and establish links and stay online
QVariant DapCommandList::networkGoTo(const QVariant &args)
{
    QStringList params = args.toStringList();

//    QString target = params[1] == "true" ? "online" : "offline";

    auto command = QString("net -net %1 go %2")
                       .arg(params[0]) // net
                       .arg(params[1]);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

/// Find and establish links and stay online
QVariant DapCommandList::networkSingleSync(const QString &net)
{
    auto command = QString("net -net %1 go sync")
                       .arg(net);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

/// Stop application and exit
void DapCommandList::restartNode()
{

    auto command = QString("exit");

    requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

QVariant DapCommandList::addNode(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("node add -net %1 -addr %2 -cell %3 -ipv4 %4 -port %5")
                       .arg(params[0])              // net
                       .arg(params[1])              // nodeAddr
                       .arg("0x0000000000000001")   // cell
                       .arg(params[2])              // ip
                       .arg(params[3]);             // port

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

QVariant DapCommandList::addNodeCommand(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("node add -net %1")
                       .arg(params[0]);              // net

    if(params.size() > 1)
    {
        command += QString(" -port %1").arg(params[1]);// port
    }

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

QVariant DapCommandList::delNode(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("node del -net %1")
                       .arg(params[0]) // net
                       .arg(params[1]) // addr node
                       ;

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

/// Delete chain files or global database
QVariant DapCommandList::removeChainsOrGdb(const QVariant &args)
{
    QString flag = args.toStringList().first();

    QString command;
    QString result;

    if(flag == "chains")
    {
        command = QString("remove -chains -all");
        result = requestToNode(QPair<QString*, QString>(m_cliPath,command)).toString();
    }
    else if(flag == "gdb")
    {
        command = QString("remove -gdb");
        result = requestToNode(QPair<QString*, QString>(m_cliPath,command)).toString();
    }
    else
    {
        command = QString("remove -chains -all");
        result = requestToNode(QPair<QString*, QString>(m_cliPath,command)).toString();

        result += "\n";

        command = QString("remove -gdb");
        result += requestToNode(QPair<QString*, QString>(m_cliPath,command)).toString();
    }

    return result;
}

/// Send a response to the client
QVariant DapCommandList::runCmd(const QVariant &args)
{
    QStringList params = args.toStringList();

    QString *nodePath = params[2].toInt() > 0 ? m_cliPath : m_toolPath;

    QString command = QString(params[0]);

    QString result = requestToNode(QPair<QString*, QString>(nodePath,command)).toString();

    if(params[1] == "isConsole")
    {
        if(result.length() > 14000)
        {
            int deleteLen = result.length() - 14000;
            result.remove(0,deleteLen);
        }
    }

    return result;
}

/// Token declaration
QVariant DapCommandList::tokenDecl(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("token_decl -net %1 -token %2 -total_supply %3 -certs %4 -signs_total 1 -signs_emission 1 -decimals 18 -type CF20")
                       .arg(params[1]) // net
                       .arg(params[2]) // name
                       .arg(params[0]) // supply
                       .arg(params[3]); // cert

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

// Не проверено
/// Token emission
QVariant DapCommandList::tokenEmission(const QVariant &args)
{
    QStringList params = args.toStringList();

    QString command = QString("token_emit -net %1 -addr %2 -token %3 -certs %4 -emission_value %5")
                          .arg(params[2]) // net
                          .arg(params[1]) // addr
                          .arg(params[3]) // name
                          .arg(params[4]) // cert
                          .arg(params[0]); // amount

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

/// Return software version
QString DapCommandList::getVersion()
{
    QString command = QString("-version");

//    return requestToNode(command);

    QString request = requestToNode(QPair<QString*, QString>(m_nodePath,command)).toString();
    auto requestList = request.split(',');

    if(requestList.isEmpty() || requestList.size() < 2)
    {
        qWarning() << "The output of the node version information has changed";
        return {};
    }
    QString version = requestList[1].trimmed();
    if(version.isEmpty())
    {
        qWarning() << "Couldn't find out the node version";
        return {};
    }
    return version;
}

QVariant DapCommandList::nodeDumpCommand(const QString &net)
{
    auto command = QString("node dump -net %1").arg(net);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

QVariant DapCommandList::nodeListCommand(const QString &net)
{
    auto command = QString("node list -net %1").arg(net);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

/// Activate or deactivate wallet
QVariant DapCommandList::walletActivateOrDeactivate(const QVariant &args)
{

    QStringList params = args.toStringList();

    bool activate = params[1] == "activate";
    QString target = activate == true ? "activate" : "deactivate";
    QString password = activate == true ? QString(" -password %1").arg(params[2]) : "";

    auto command = QString("wallet %1 -w %2%3")
                       .arg(target) // "activate" or "deactivate"
                       .arg(params[0]) // name wallet
                       .arg(password); // password

    if(activate) command += QString(" -ttl %1").arg(params[3]);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

/// Create a new order and tx with specified amount of datoshi to exchange with specified rate (buy / sell)
/// srv_xchange order create -net <net_name> -token_sell <token_ticker> -token_buy <token_ticker> -w <wallet_name> -value <value> -rate <value> -fee <value>
QVariant DapCommandList::xchangeOrderCreate(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("srv_xchange order create -net %1 -token_sell %2 -token_buy %3 -w %4 -value %5 -rate %6 -fee %7")
                       .arg(params[0]) // net
                       .arg(params[1]) // token sell
                       .arg(params[2]) // token buy
                       .arg(params[3]) // wallet name
                       .arg(params[4]) // amount
                       .arg(params[5]) // rate
                       .arg(params[6]);// fee

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::txCondCreate(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("tx_cond_create -net %1 -token %2 -w %3 -cert %4 -value %5 -fee %6 -unit %7 -srv_uid %8")
                       .arg(params[0]) // net
                       .arg(params[1]) // token
                       .arg(params[2]) // wallet name
                       .arg(params[3]) // cert
                       .arg(params[4]) // value
                       .arg(params[5]) // fee
                       .arg(params[6]) // unit
                       .arg(params[7]);// srv uid

    return requestToNode(QPair<QString*, QString>(m_cliPath,command),Dap::RequestType::JSON);
}

QVariant DapCommandList::stakeLockHold(const QVariant &args)
{
    QStringList params = args.toStringList();

    QString command = QString("stake_lock hold -net %1 -w %2 -time_staking %3 -token %4 -value %5 -reinvest %6 -fee %7")
                        .arg(params[0]) // net
                        .arg(params[1]) // wallet
                        .arg(params[2]) // time
                        .arg(params[3]) // token
                        .arg(params[4]) // value
                        .arg(params[5]) // reinvest
                        .arg(params[6]); // fee

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

QVariant DapCommandList::stakeLockTake(const QVariant &args)
{
    QStringList params = args.toStringList();

    QString command = QString("stake_lock take -net %1 -w %2 -tx %3 -fee %4")
                        .arg(params[0]) // net
                        .arg(params[1]) // wallet
                        .arg(params[2]) // hash
                        .arg(params[3]); // fee

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

QVariant DapCommandList::historyInfoByHash(const QVariant &args)
{
    QStringList params = args.toStringList();

    QString command = QString("tx_history -net %1 -tx %2")
                        .arg(params[0]) // net
                        .arg(params[1]); // tx

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

/// Exchange tokens with specified order within specified net name. Specify how many datoshies to sell with rate specified by order
///srv_xchange purchase -order <order hash> -net <net_name> -w <wallet_name> -value <value> -fee <value>
QVariant DapCommandList::xchangeOrderPurchase(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("srv_xchange purchase -order %1 -net %2 -w %3 -value %4 -fee %5")
                       .arg(params[0]) // order hash
                       .arg(params[1]) // net
                       .arg(params[2]) // wallet name
                       .arg(params[3]) // coins
                       .arg(params[4]);// fee

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

/// Remove order with specified order hash in specified net name
QVariant DapCommandList::xchangeOrderRemove(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("srv_xchange order remove -net %1 -order %2 -w %3 -fee %4")
                       .arg(params[0]) // net
                       .arg(params[1]) // hash
                       .arg(params[2]) // wallet
                       .arg(params[3]); // fee
    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

/// List mempool (entries or transaction) for (selected chain network or wallet)
QVariant DapCommandList::getListMempool(const QString &net, const QString &addr, const QString &chain)
{
    auto command = QString("mempool list -net %1")
                       .arg(net);

    if(!addr.isEmpty())
        command += QString(" -addr %1").arg(addr);
    if(!chain.isEmpty())
        command += QString(" -chain %1").arg(chain);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::getMempoolHashInfo(const QVariant &args)
{
    QStringList params = args.toStringList();
    if(params.isEmpty())
    {
        qWarning() << "Not arguments";
        return {};
    }
    auto command = QString("mempool check -net %1 -datum %2")
                       .arg(params[0])
                       .arg(params[1]);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::getWalletAddress(const QString &net, const QString &wallet)
{
    auto result = getWalletInfo(net, wallet);

    return result.address;
}

QVariant DapCommandList::createVPNOrder(const QVariant &args)
{
    QStringList params = args.toStringList();
    auto command = QString("net_srv");
    QString tmpParam;
    if(getParamForPairList(params, "network", tmpParam)) command += QString(" -net %1 order create").arg(tmpParam);
    else return QString("The network parameter was not found.");

    if(getParamForPairList(params, "direction", tmpParam)) command += QString(" -direction %1").arg(tmpParam);
    else return QString("The direction parameter was not found.");

    if(getParamForPairList(params, "srv_uid", tmpParam)) command += QString(" -srv_uid %1").arg(tmpParam);
    else return QString("The srv_uid parameter was not found.");

    if(getParamForPairList(params, "price", tmpParam)) command += QString(" -price %1").arg(tmpParam);
    else return QString("The price parameter was not found.");

    if(getParamForPairList(params, "price_unit", tmpParam)) command += QString(" -price_unit %1").arg(tmpParam);
    else return QString("The price_unit parameter was not found.");

    if(getParamForPairList(params, "price_token", tmpParam)) command += QString(" -price_token %1").arg(tmpParam);
    else return QString("The price_token parameter was not found.");

    if(getParamForPairList(params, "units", tmpParam)) command += QString(" -units %1").arg(tmpParam);
    else return QString("The units parameter was not found.");

    if(getParamForPairList(params, "cert", tmpParam)) command += QString(" -cert %1").arg(tmpParam);
    else return QString("The cert parameter was not found.");

    if(getParamForPairList(params, "region", tmpParam)) command += QString(" -region %1").arg(tmpParam);

    if(getParamForPairList(params, "continent", tmpParam)) command += QString(" -continent %1").arg(tmpParam);

    if(getParamForPairList(params, "node_addr", tmpParam)) command += QString(" -node_addr %1").arg(tmpParam);

    if(getParamForPairList(params, "tx_cond", tmpParam)) command += QString(" -tx_cond %1").arg(tmpParam);

    if(getParamForPairList(params, "expires", tmpParam)) command += QString(" -expires %1").arg(tmpParam);

    if(getParamForPairList(params, "ext", tmpParam)) command += QString(" -ext %1").arg(tmpParam);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

QVariant DapCommandList::createStakeOrder(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("srv_stake order create -net %1 -value %2 -cert %3")
                       .arg(params[0])   // network
                       .arg(params[1])   // value
                       .arg(params[2]);  // cert

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::createOrderValidator(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("srv_stake order create validator -net %1 -value_min %2 -value_max %3 -tax %4 -cert %5")
                       .arg(params[0])   // network
                       .arg(params[1])   // minimum_stake_value
                       .arg(params[2])   // maximum_stake_value
                       .arg(params[3])   // percent
                       .arg(params[4]);  // cert

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::createOrderStaker(const QVariant &args)
{
    QStringList params = args.toStringList();

    auto command = QString("srv_stake order create staker -net %1 -w %2 -value %3 -fee %4 -tax %5")
                       .arg(params[0])   // network
                       .arg(params[1])   // wallet name
                       .arg(params[2])   // value
                       .arg(params[4])   // value fee validator
                       .arg(params[3]);  // tax

    if(params.size() > 5 && (params[5] != "web3" &&  params[5] != "feeAdded" && params[5] != "queueHash"))
    {
        command.append(QString(" -addr %1").arg(params[5]));
    }

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::getListKeysCommand(const QString &net)
{
    QString command = QString("srv_stake list keys -net %1").arg(net);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::getListChains()
{
    auto command = QString("net list chains");

    return requestToNode(QPair<QString*, QString>(m_cliPath,command));
}

QVariant DapCommandList::votingCreate(const QVariant &args)
{
    QStringList params = args.toStringList();
    auto command = QString("poll create ");
    QString tmpParam;
    if(getParamForPairList(params, "net", tmpParam)) command += QString(" -net %1").arg(tmpParam);
    else return QString("The network parameter was not found.");

    if(getParamForPairList(params, "walletName", tmpParam)) command += QString(" -w %1").arg(tmpParam);
    else return QString("The walletName parameter was not found.");

    if(getParamForPairList(params, "question", tmpParam)) command += QString(" -question %1").arg(tmpParam);
    else return QString("The question parameter was not found.");

    if(getParamForPairList(params, "voteOptions", tmpParam)) command += QString(" -options %1").arg(tmpParam);
    else return QString("The voteOptions parameter was not found.");

    if(getParamForPairList(params, "fee", tmpParam)) command += QString(" -fee %1").arg(tmpParam);
    else return QString("The fee parameter was not found.");

    if(getParamForPairList(params, "expire", tmpParam)) command += QString(" -expire %1").arg(tmpParam);

    if(getParamForPairList(params, "maxVotesCount", tmpParam)) command += QString(" -max_votes_count %1").arg(tmpParam);

    if(getParamForPairList(params, "delegatedKeyRequired", tmpParam)) command += QString(" -delegated_key_required ");

    if(getParamForPairList(params, "voteChangingAllowed", tmpParam)) command += QString(" -vote_changing_allowed ");

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::votingVote(const QVariant &args)
{
    QStringList params = args.toStringList();
    auto command = QString("poll vote ");
    QString tmpParam;
    if(getParamForPairList(params, "net", tmpParam)) command += QString(" -net %1").arg(tmpParam);
    else return QString("The network parameter was not found.");

    if(getParamForPairList(params, "walletName", tmpParam)) command += QString(" -w %1").arg(tmpParam);
    else return QString("The walletName parameter was not found.");

    if(getParamForPairList(params, "hash", tmpParam)) command += QString(" -hash %1").arg(tmpParam);
    else return QString("The question parameter was not found.");

    if(getParamForPairList(params, "optionIdx", tmpParam)) command += QString(" -option_idx %1").arg(tmpParam);
    else return QString("The voteOptions parameter was not found.");

    if(getParamForPairList(params, "fee", tmpParam)) command += QString(" -fee %1").arg(tmpParam);
    else return QString("The fee parameter was not found.");

    if(getParamForPairList(params, "cert", tmpParam)) command += QString(" -cert %1").arg(tmpParam);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command), Dap::RequestType::JSON);
}

QVariant DapCommandList::votingList(const QVariant &args)
{
    QStringList params = args.toStringList();
    if(params.isEmpty())
    {
        return {};
    }
    auto command = QString("poll list -net %1").arg(params[0]);

    return requestToNode(QPair<QString*, QString>(m_cliPath,command),Dap::RequestType::JSON);
}

QVariant DapCommandList::votingDump(const QVariant &args)
{
    QStringList params = args.toStringList();
    if(params.isEmpty() || params.size() < 2)
    {
        return {};
    }
    auto command = QString("poll dump -net %1 -hash %2").arg(params[0]).arg(params[1]);

    return requestToNode(QPair<QString*, QString>(m_cliPath, command), Dap::RequestType::JSON);
}

// Convert old 'cli-rpc' cmd to new 'json-rpc'
QString DapCommandList::convertCliToJson(const QString &cmd) {

    QStringList cmdStringList = cmd.split(" ");

    QString cmdMethod = cmdStringList[0];
    QString cmdID = "100"; // temp

    if(!cmdMethod.isEmpty()) {

        QJsonObject jsonObj;
        QJsonArray jsonArr;
        QString jsonRequest;

        // method
        jsonObj["method"] = cmdMethod;

        // params
        QString jsonParams = cmdMethod;
        for(int i = 1 ; i < cmdStringList.size() ; i++ ) {
            jsonParams.append(";");
            jsonParams.append(cmdStringList[i]);
        }
        jsonArr.append(QJsonValue(jsonParams));
        jsonObj["params"] = jsonArr;

        // id
        jsonObj["id"] = cmdID;

        QJsonDocument jsonDoc(jsonObj);
        jsonRequest = QString::fromUtf8(jsonDoc.toJson());

        return jsonRequest;
    } else {
        qWarning() << "Unknown command: " << cmd;
        return "";
    }
}

QVariant DapCommandList::wrapJsonError(const QString& msg)
{
    return QJsonDocument::fromJson("{\"error\": \"" + msg.toUtf8() + "\"}");
}

QVariant DapCommandList::requestToNodeJSON(QPair<QString*, QString> cmd)
{
    // Convert old 'cli-rpc' cmd to new 'json-rpc'
    QString jsonRequest = convertCliToJson(cmd.second);

    if(jsonRequest.isEmpty()) return wrapJsonError("command is empty");

    // HTTP request
    QString answer{""};
    QString httpRequest = "POST http://localhost/connect HTTP/1.1\r\n"
                          "Host: localhost\r\n"
                          "Content-Type: application/json\r\n"
                          "Content-Length: " + QString::number(jsonRequest.length()) + "\r\n"
                                                                    "\r\n" + jsonRequest;


//    if(m_socketPath.isEmpty())
//    {
        updateConfigParams();
//    }

    QIODevice * socket;

    if(m_socketPort)
    {
        socket = new QTcpSocket();
        ((QTcpSocket*)socket)->connectToHost("127.0.0.1", m_socketPort);

        if(!((QTcpSocket*)socket)->waitForConnected(60000))
        {
            // connecting error
            qWarning() << "Connection to 127.0.0.1 failed:" << socket->errorString() << ". Path: " << m_socketPath;
            return wrapJsonError("connection to tcp socket failed");
        }
    }
#ifdef Q_OS_UNIX
    else if(!m_socketPath.isEmpty())
    {
        // Linux & Mac (unix-socket)
        socket = new QLocalSocket();
        ((QLocalSocket*)socket)->setServerName(m_socketPath);
        ((QLocalSocket*)socket)->connectToServer();

        if (!((QLocalSocket*)socket)->waitForConnected(60000))
        {
            // connecting error
            qWarning() << "Connection to 127.0.0.1 failed:" << socket->errorString() << ". Path: " << m_socketPath;
            return wrapJsonError("connection to tcp socket failed");
        }
    }
#endif
    else
    {
        qWarning()<<"Error init JSON-RPC connect params";
        return wrapJsonError("Error init JSON-RPC connect params");
    }

    // send request
    socket->write(httpRequest.toUtf8());
    // read reply
    if (socket->waitForReadyRead(180000))
    {
        QByteArray responseData = socket->readAll();

        //0-header
        //1-body
        QStringList dataRcv = QString::fromUtf8(responseData).split("\r\n\r\n");

        if(dataRcv.count() > 1)
        {
            QStringList headerSplit = dataRcv.first().split("Content-Length: ");
            QString splitItem = headerSplit.last();
            if(splitItem.contains("Processing-Time: "))
                splitItem = splitItem.split("Processing-Time: ").first().simplified();

            int contentSize = splitItem.toInt();
            answer = dataRcv.last(); //write first body

            while (contentSize != answer.size())
            {
                ((QAbstractSocket*)socket)->waitForReadyRead();
                if(!socket->bytesAvailable())
                    break;

                QByteArray buff = socket->readAll();
                answer += QString::fromUtf8(buff);
            }
        }
        else
            qDebug() << "Incorrect dataRcv count";
    }
    else
        qWarning() << "Error. Can't read answer from tcp socket:" << socket->errorString();


    socket->deleteLater();

    //FOR DEBUG
   // std::cout<< "---------- " << QTime::currentTime().toString().toStdString() << " ----------" << std::endl;
   // std::cout<< QString("JSON REQ ").toStdString() << jsonRequest.toStdString() << std::endl;
   // std::cout<< QString("JSON REP ").toStdString() << answer.toStdString() << std::endl;

    if (answer.isEmpty()) {
        qWarning() << "Can't parsing body of HTTP answer, because not found";
        return  wrapJsonError("can't parsing body of HTTP answer, because \r\n\r\n not found");
    }

    // Parsing json answer
    QJsonDocument jsonDoc = QJsonDocument::fromJson(answer.toUtf8());
    if (!jsonDoc.isObject()) {
        qWarning() << "Can't parsing answer, because it's not a JSON object";
        return wrapJsonError("can't parsing answer, because it's not a JSON object");
    }

    return QVariant(jsonDoc);
}

bool DapCommandList::checkContentLenght(QString body)
{
    if(body.indexOf("Content-Length: ") < 0 || body.indexOf("\r\n\r\n") < 0)
    {
        qDebug() << "Not found Content-Length and content." << body;
        return true;
    }
    QString temp = body.split("Content-Length: ").last();
    bool ok;
    uint32_t contentLength = temp.split("\r\n").first().toUInt(&ok);
    if(!ok)
    {
        qDebug() << "Can't parsing content length." << body;
        return true;
    }
    QString content = body.split("\r\n\r\n").last();

    return static_cast<uint32_t>(content.length()) >= contentLength;
}


QVariant DapCommandList::requestToNodeCLI(QPair<QString*, QString> cmd)
{
    QStringList args = cmd.second.split(" ");
    QProcess process;
    process.setProgram(*cmd.first);
    process.setArguments(args);
    process.start();
    /// TODO: if you need a list of output commands, then uncomment.
    /// In the usual case, it is better to avoid this because it causes logging redundancy.
    if(args.contains("-password"))
    {
        int idx = args.indexOf("-password");
        if(args.count() >= idx+1)
            args[idx+1] = "****";
    }

    qDebug()<< process.program() << "command:" << args;

    int attempt_count = 5;

    while(attempt_count)
    {
        if(process.error() == QProcess::FailedToStart)
            process.start();

        if(process.waitForFinished(120000)) { //2 min
            auto result = QString::fromLatin1(process.readAll());
            return result;
        }

        attempt_count--;
    }

    qWarning() << "Timeout request: " << args << QString("Error string: %2. Error number: %3.").arg(process.errorString()).arg(process.error());
    return QString();
}

QVariant DapCommandList::requestToNode(QPair<QString*, QString> cmd, Dap::RequestType type)
{
    return type == Dap::RequestType::CLI ? requestToNodeCLI(cmd) : requestToNodeJSON(cmd);
}

Dap::JsonCommandsRespond DapCommandList::getJsonResult(const QVariant& resultVar, const QString& command)
{
    Dap::JsonCommandsRespond respond;
    QJsonDocument resultDocument = resultVar.toJsonDocument();
    if(resultDocument.isNull())
    {
        qWarning() << "Couldn't get JsonDocument (" << command << ")";
        return {};
    }
    QJsonObject resultObject = resultDocument.object();
    QString error{""};
    if(resultObject[Dap::RESULT_TEXT_JSON].toArray().first().toObject().contains(Dap::ERROR_TEXT_JSON))
    {
        auto errorObjectArray = resultObject[Dap::RESULT_TEXT_JSON].toArray().first().toObject()[Dap::ERROR_TEXT_JSON].toArray().first().toObject();
        if(!errorObjectArray.isEmpty())
        {
            QString errorCode{""};
            QString errorMessage{""};

            QString resultError{""};

            if(errorObjectArray.contains("code"))
            {
                errorCode = errorObjectArray["code"].toString();
                resultError = resultError + " Code: " + errorCode;
            }
            else
            {
                qWarning() << "[DapCommandList] [getJsonResult] Changed the error output for command --"
                           << command << "-- The -code- element was not found";
            }

            if(errorObjectArray.contains("message"))
            {
                errorMessage = QString::number(errorObjectArray["message"].toInt());
                resultError = resultError + " Message: " + errorMessage;
            }
            else
            {
                qWarning() << "[DapCommandList] [getJsonResult] Changed the error output for command --"
                           << command << "-- The -message- element was not found";
            }

            if(resultError.isEmpty())
            {
                error = "There have been changes in the responses from the node.";
                qWarning() << "[DapCommandList] [getJsonResult] Changed the error output for command --"
                           << command << "-- The null element does not contain the error element.";
            }
            else
            {
                error = resultError;
            }
        }
        else
        {
            error = "There have been changes in the responses from the node.";
            qWarning() << "[DapCommandList] [getJsonResult] Changed the error output for command --"
                       << command << "-- The array is empty.";
        }
    }
    respond.result = resultObject[Dap::RESULT_TEXT_JSON];
    respond.error = error;
    return respond;
}

void DapCommandList::updateConfigParams()
{
//    qDebug() << "[updateConfigParams] An attempt to get data from the config.";

    QString nodePath, listenPathCfg, listenAddrCfg;

    nodePath = DapNodePathManager::getInstance().nodePaths.nodeConfigPath;
    listenPathCfg = DapConfigToolController::getInstance().getConfigParam("cellframe-node","cli-server", "listen-path");
    listenAddrCfg = DapConfigToolController::getInstance().getConfigParam("cellframe-node","cli-server", "listen-address");

    listenAddrCfg = DapConfigToolController::getInstance().parseArrayConfigParam(listenAddrCfg).first();

#ifdef Q_OS_UNIX
    if(!nodePath.isEmpty())
#else
    if(!listenAddrCfg.isEmpty() || listenAddrCfg.contains(":"))
#endif
    {
        QString newPath = nodePath + DapConfigToolController::getInstance().parseArrayConfigParam(listenPathCfg).first();
        uint16_t newPort = (listenAddrCfg.isEmpty() || !listenAddrCfg.contains(":")) ? 0 : listenAddrCfg.split(":").last().toUInt();

        if(newPath != m_socketPath || newPort != m_socketPort)
        {
            qDebug()<<"[DapCommandList::updateConfigParams]  Update JSON RPC connect params";
            qDebug()<<"Old path = " << m_socketPath << " Old port = " << m_socketPort;
            qDebug()<<"New path = " << newPath      << " New port = " << newPort;

            m_socketPath = newPath;
            m_socketPort = newPort;
        }
    }
}

bool DapCommandList::getParamForPairList(const QStringList& list, const QString& paramName, QString& param)
{
    if(list.contains(paramName))
    {
        int index = list.indexOf(paramName);
        if(list.size() < index + 2)
        {
            qWarning() << "A parameter may have been omitted: " << paramName;
            param.clear();
            return false;
        }
        if(!list[index+1].isEmpty())
        {
            param = list[index+1];
            return true;
        }
    }
    return false;
}

QString DapCommandList::getError(const QJsonValue &json)
{
    auto errObj = json.toArray()[0].toObject();
    if(!errObj.contains("error")) return QString();
    auto errMsg = errObj["error"].toObject();
    if(!errMsg.contains("message")) return QString();
    return errMsg["message"].toString();
}

bool DapCommandList::testJsonBlock(bool expression, const QString& message, const QString& commentForLog)
{
    if(expression)
    {
        qWarning() << QString("[%1] The output of the node has changed or some kind of error has occurred.(%2)").arg(commentForLog).arg(message);
        return false;
    }
    return true;
}
