#include "DapDictionaryCommand.h"
#include <QDataStream>
#include <QByteArray>

DapDictionaryCommand::DapDictionaryCommand(const QString &asServiceName, QObject *parent, const QString &asCliPath)
    : DapAbstractCommand{asServiceName, parent, asCliPath}
    , m_dictionary(new HelpDictionaryServiceController)
{
}

void DapDictionaryCommand::updateVersionNode(bool isUpdate)
{
    QString version = getVersion();
    if(!version.isEmpty() && version != m_dictionary->getNodeVersion())
    {
        m_dictionary->setNodeVersion(std::move(version), isUpdate);
    }
}

void DapDictionaryCommand::updateData()
{
    updateCerts();
    updateTokens();
    updateMempoolTokens();
    updateChains();
}

DapDictionaryCommand::~DapDictionaryCommand()
{
    delete m_dictionary;
}

void DapDictionaryCommand::initRegularRequestController()
{
    if(m_regularController)
    {
        connect(m_regularController, &DapRegularRequestsController::versionNodeUpadeted, this, &DapDictionaryCommand::slotUpdateVersionNode);
    }
    slotUpdateVersionNode();
}

void DapDictionaryCommand::slotUpdateVersionNode()
{
    updateVersionNode();
}

QVariant DapDictionaryCommand::respondToClient(const QVariant &args)
{
    if(!m_dictionary)
    {
        return {};
    }
    QJsonDocument result;
    QStringList params = args.toStringList();

    auto getJsonWithHeader = [](const QString& head, const QJsonDocument& doc, const QString& versionNode = "") -> QJsonDocument
    {
        QJsonObject result;
        result["head"] = head;
        result["data"] = doc.object();
        if(!versionNode.isEmpty())
        {
            result["version"] = versionNode;
        }

        return QJsonDocument(result);
    };

    if(params[0] == "getNewDictionary")
    {
        updateVersionNode(false);
        result = getJsonWithHeader("dictionary", m_dictionary->getNewJsonDictionary(), m_dictionary->getNodeVersion());
    }
    else if (params[0] == "getDictionary")
    {
        updateVersionNode(false);
        result = getJsonWithHeader("dictionary", m_dictionary->getJsonDictionary(), m_dictionary->getNodeVersion());
    }
    else if(params[0] == "getData")
    {
        result = getJsonWithHeader("data", getData());
    }
    else if(params[0] == "updateData")
    {
        updateData();
        result = getJsonWithHeader("data", getData());
    }

    QByteArray resByte;
    QDataStream streamer(&resByte, QIODevice::WriteOnly);
    streamer << result;

    return resByte.toHex();
}

QJsonDocument DapDictionaryCommand::getData()
{
    auto stringListToJson = [](const QStringList& list) -> QJsonArray
    {
       QJsonArray jsonArray;
       for(const QString& str: list)
       {
           jsonArray.append(str);
       }
       return jsonArray;
    };

    QJsonObject data;
    data["wallets"] = stringListToJson(getListWallets().keys());
    data["networks"] = stringListToJson(getListNetworks());
    data["pubCertName"] = stringListToJson(m_pubCertName);
    data["privCertName"] = stringListToJson(m_privCertName);

    auto mapStrLToJson = [&stringListToJson](const QMap<QString, QStringList>& chains) -> QJsonArray
    {
        QJsonArray jsonArray;
        QMapIterator<QString, QStringList> it(chains);
        while(it.hasNext())
        {
            it.next();
            QString key = it.key();
            QStringList list = it.value();
            QJsonArray jsonList = stringListToJson(list);
            QJsonObject tmpObject;
            tmpObject["key"] = key;
            tmpObject["data"] = jsonList;
            jsonArray.append(std::move(tmpObject));
        }
        return jsonArray;
    };

    data["chains"] = mapStrLToJson(m_chains);
    data["mempoolTokensTicker"] = mapStrLToJson(m_mempoolTokensTicker);

    auto mapTokenToJson = [&stringListToJson](const QMap<QPair<QString, QString>, QStringList>& tokensTicker) -> QJsonArray
    {
        QJsonArray jsonArray;
        QMapIterator<QPair<QString, QString>, QStringList> it(tokensTicker);
        while(it.hasNext())
        {
            it.next();
            QPair<QString, QString> key = it.key();
            QStringList list = it.value();
            QJsonArray jsonList = stringListToJson(list);
            QJsonObject tmpObject;
            tmpObject["wallet"] = key.first;
            tmpObject["network"] = key.second;
            tmpObject["data"] = jsonList;
            jsonArray.append(std::move(tmpObject));
        }
        return jsonArray;
    };

    data["tokensTicker"] = mapTokenToJson(m_tokensTicker);
    return QJsonDocument(data);
}

void DapDictionaryCommand::updateCerts()
{
    m_pubCertName.clear();
    m_privCertName.clear();
    QDir pubDir(PUB_CERT_PATH);
    pubDir.setFilter(QDir::Files | QDir::Hidden);
    QStringList files = pubDir.entryList();
    for (int i = 0; i < files.length(); ++i)
    {
        QString s = files[i].remove(".dcert");
        if (s != "." && s != "..")
            m_pubCertName.append(s);
    }

    QDir privDir(PRIV_CERT_PATH);
    privDir.setFilter(QDir::Files | QDir::Hidden);
    files = privDir.entryList();
    for (int i = 0; i < files.length(); ++i)
    {
        QString s = files[i].remove(".dcert");
        if (s != "." && s != "..")
            m_privCertName.append(s);
    }
}

void DapDictionaryCommand::updateTokens()
{
    QMap<QPair<QString, QString>, QStringList> result;

    QMap <QString, QString> wallets = getListWallets();
    QStringList nets = getListNetworks();
    for(const QString& wallet: wallets.keys())
    {
        for(const QString& net: nets)
        {
            auto resultRequest = cmdList->getWalletInfo(net,wallet);
            QStringList listTicker;
            for(const auto& token: resultRequest.tokenList)
            {
                listTicker.append(token.token);
            }

            if(!listTicker.isEmpty())
            {
                result.insert({wallet, net}, listTicker);
            }
        }
    }
    if(!result.isEmpty())
    {
        m_tokensTicker = std::move(result);
    }
}

void DapDictionaryCommand::updateMempoolTokens()
{
    QMap<QString, QStringList> result;
    QStringList nets = getListNetworks();

    for(const QString& net: nets)
    {
        auto resultRequest = cmdList->getListTokens(net);

        QStringList listTicker = resultRequest.keys();

        if(!listTicker.isEmpty())
        {
            result.insert(net, listTicker);
        }
    }
    if(!result.isEmpty())
    {
        m_mempoolTokensTicker = std::move(result);
    }
    else
    {
        qWarning() << "The search for Tickers in the mempool yielded no results";
    }
}

void DapDictionaryCommand::updateChains()
{
    QMap<QString, QStringList> result;
    QString resultRequest = cmdList->getListChains().toString();

    resultRequest = resultRequest.remove("Networks:\n\t");
    resultRequest = resultRequest.remove(":");
    resultRequest = resultRequest.remove("\n");
    resultRequest = resultRequest.remove("\r");
    QStringList list = resultRequest.split("\t");
    list.removeAll("");

    if (list.isEmpty())
        return;

    QString net;
    QStringList tmpList;
    auto addToResult = [&result, &net, &tmpList]()
    {
        if(!net.isEmpty() && !tmpList.isEmpty())
        {
            result.insert(net, tmpList);
            net.clear();
            tmpList.clear();
        }
    };

    for (int i = 0; i < list.length(); ++i)
    {
        if(getListNetworks().contains(list[i]))
        {
            addToResult();
            net = list[i];
        }
        else if(!list[i].contains('|'))
        {
            tmpList.append(list[i]);
        }
    }
    addToResult();

    if(!result.isEmpty())
    {
        m_chains = std::move(result);
    }
}

