#include "DapGetNetworksStateCommand.h"

const QMap<QString, QString> DapGetNetworksStateCommand::s_stateStrings = {
    { "NET_STATE_OFFLINE", "OFFLINE"},
    { "NET_STATE_ONLINE", "ONLINE"},
    { "NET_STATE_LINKS_PREPARE", "LINKS PREPARE"},
    { "NET_STATE_LINKS_ESTABLISHED", "LINKS ESTABLISHED"},
    { "NET_STATE_LINKS_CONNECTING", "LINKS CONNECTING"},
    { "NET_STATE_SYNC_CHAINS", "SYNC CHAINS"},
    { "NET_STATE_SYNC_GDB", "SYNC GDB"},
    { "NET_STATE_ADDR_REQUEST", "ADDR REQUEST"},
    { "UNDEFINED", "ERROR"}
};

/// Overloaded constructor.
/// @param asServiceName Service name.
/// @param parent Parent.
/// @details The parent must be either DapRPCSocket or DapRPCLocalServer.
/// @param asCliPath The path to cli nodes.
DapGetNetworksStateCommand::DapGetNetworksStateCommand(const QString &asServicename, QObject *parent, const QString &asCliPath)
    : DapAbstractCommand(asServicename, parent, asCliPath)
{
}

/// Send a response to the client.
/// @details Performed on the service side.
/// @param arg1...arg10 Parameters.
/// @return Reply to client.
QVariant DapGetNetworksStateCommand::respondToClient(const QVariant &args)
{
    DapAbstractCommand::respondToClient(args);

    QStringList params = args.toStringList();

    bool isWeb3 = params.contains(WEB3_KEY);
    NetworkSpace::NetworkStatus w3Result;

    QStringList listNetworks;
    QString errorMsg;
    QJsonArray netArr;

    if(isWeb3)
    {
        listNetworks.append(params.first());
    }
    else
    {
        listNetworks = getListNetworks();
    }

    // Get state of each network
    for(const QString &net : listNetworks)
    {
        QJsonObject netOjb;
        auto result = cmdList->getNetworkState(net, errorMsg);

        if(!errorMsg.isEmpty())
            qWarning() << "Error. There are problems in the output." << errorMsg;

        netOjb.insert("name", result.network.isEmpty() ? net : result.network);
        netOjb.insert("networkState", result.currantState);
        netOjb.insert("targetState", result.targetState);
        netOjb.insert("nodeAddress", result.currentAddress);
        netOjb.insert("errorMessage", errorMsg);
        netOjb.insert("linksCount", result.totalLinks.isEmpty() ? "0" : result.totalLinks);
        netOjb.insert("activeLinksCount", result.activeLinks.isEmpty() ? "0" : result.activeLinks);
        netOjb.insert("syncPercent", result.syncPercent);
        netOjb.insert("displayNetworkState", convertState(result.currantState));
        netOjb.insert("displayTargetState", convertState(result.targetState));
        netArr.append(netOjb);

        if(isWeb3)
            w3Result = result;
    }

    QJsonObject resultObj;

    if(isWeb3)
    {
        if(!errorMsg.isEmpty())
            resultObj.insert(ERROR_KEY, errorMsg);

        QJsonObject sourceData = netArr[0].toObject();
        QJsonObject data;
        data.insert("active_links_count", w3Result.activeLinks);
        data.insert("links_count", w3Result.totalLinks);
        data.insert("node_address", w3Result.currentAddress);
        data.insert("state", w3Result.currantState);
        data.insert("target_state", w3Result.targetState);
        data.insert("sync_percent", w3Result.syncPercent);

        resultObj.insert(RESULT_KEY, data);
        addWeb3Result(resultObj, args);
    }
    else
    {
        resultObj.insert(RESULT_KEY, netArr);
    }

    QJsonDocument resultDoc(resultObj);
    return resultDoc.toJson();
}

void DapGetNetworksStateCommand::replyFromService()
{
    DapRpcServiceReply *reply = static_cast<DapRpcServiceReply *>(sender());
    QVariant replyVariant = reply->response().toJsonValue().toVariant();

    QJsonDocument replyDoc = QJsonDocument::fromJson(replyVariant.toByteArray());
    QJsonObject replyObj = replyDoc.object();

    if(replyObj[RESULT_KEY].isArray())
    {
        QJsonArray resultArr = replyObj[RESULT_KEY].toArray();
        QJsonDocument resultDoc(resultArr);
        emit serviceResponded(resultDoc.toJson());
    }
    else
    {
        emit serviceResponded({});
    }

}

QString DapGetNetworksStateCommand::convertState(const QString &state)
{
    if(s_stateStrings.contains(state))
    {
        return s_stateStrings[state];
    }

    auto substr = QString("NET_STATE_");
    if(state.contains(substr))
    {
        QString result = state;
        result.remove(substr);
        return result;
    }

    return state;
}

