#include "DapGetWalletInfoCommand.h"
#include "../DapTypes/DapCoin.h"

/// Overloaded constructor.
/// @param asServiceName Service name.
/// @param parent Parent.
/// @details The parent must be either DapRPCSocket or DapRPCLocalServer.
/// @param asCliPath The path to cli nodes.
DapGetWalletInfoCommand::DapGetWalletInfoCommand(const QString &asServicename, QObject *parent, const QString &asCliPath)
    : DapAbstractCommand(asServicename, parent, asCliPath)
{
}

/// Send a response to the client.
/// @details Performed on the service side.
/// @param arg1...arg10 Parameters.
/// @return Reply to client.
QVariant DapGetWalletInfoCommand::respondToClient(const QVariant &args)
{
    DapAbstractCommand::respondToClient(args);

    QStringList params = args.toStringList();
    bool isWeb3 = params.contains(WEB3_KEY);
    QString error;
    QJsonObject walletObj;

    QString walletName = params[0];

    // get list networks
    QStringList netlist = getListNetworks();

    // get list wallets
    auto res = getListWallets();

    // check status wallet
    QString status;

    if(res.contains(walletName))
    {
        status = res[walletName];
        walletObj.insert("status", status);
    }
    else
    {
        qWarning() << "Unexpected behavior. In the list of wallets not searched for. walletName = " << walletName;
        QJsonDocument resultDoc(QJsonObject({{RESULT_KEY, {}}}));
        return resultDoc.toJson();
    }

    walletObj.insert("name", walletName);

    //------//
    QJsonArray networkArray;
    if(status == "Active" || status.isEmpty())
    {       
        for (const QString &net: netlist)
        {
            // get wallet addresses
            auto result = cmdList->getWalletInfo(net, walletName);

            QJsonArray tokenArray;
            for(const auto& token: result.tokenList)
            {
                QJsonObject tokenObj;

                tokenObj.insert(isWeb3 ? "tokenName" :"name", token.token );
                tokenObj.insert(isWeb3 ? "balance" :"coins", token.coins);

                tokenObj.insert("datoshi", token.datoshi);

                tokenObj.insert("availableCoins", "");
                tokenObj.insert("availableDatoshi", "");
                QString availableDatoshi, availableCoins;

                if(getAvailableBalance(walletName, token.datoshi, net,
                                        token.token, availableDatoshi, availableCoins))
                {
                    tokenObj.insert("availableCoins", availableCoins);
                    tokenObj.insert("availableDatoshi", availableDatoshi);
                }

                tokenArray.append(tokenObj);
            }
            QJsonObject networkObj;
            networkObj.insert("tokens", tokenArray);
            networkObj.insert("address", result.address);
            networkObj.insert(isWeb3 ? "network" :"name", net);

            networkArray.append(networkObj);

        }
        walletObj.insert("networks", networkArray);
    }
    QJsonDocument resultDoc;
    if(isWeb3)
    {
        QJsonObject resultObject;
        resultObject.insert(RESULT_KEY, networkArray);
        addWeb3Result(resultObject, args);
        resultDoc.setObject(resultObject);
    }
    else
    {
        resultDoc.setObject(QJsonObject({{RESULT_KEY, walletObj}}));
    }
//
    return resultDoc.toJson();
}

/// Reply from service.
/// @details Performed on the service side.
/// @return Service reply.
void DapGetWalletInfoCommand::replyFromService()
{
    DapRpcServiceReply *reply = static_cast<DapRpcServiceReply *>(sender());
    QVariant replyVariant = reply->response().toJsonValue().toVariant();

    QJsonDocument replyDoc = QJsonDocument::fromJson(replyVariant.toByteArray());
    QJsonObject replyObj = replyDoc.object();

    if(replyObj[RESULT_KEY].isObject())
    {
        QJsonObject resultObj = replyObj[RESULT_KEY].toObject();
        QJsonDocument resultDoc(resultObj);
        emit serviceResponded(resultDoc.toJson());
    }
    else
    {
        QString resultStr = replyObj[RESULT_KEY].toString();
        emit serviceResponded(resultStr);
    }
}

void DapGetWalletInfoCommand::queueDataUpdate(const QVariant &data)
{
    m_queueData = data.value<QVector<Command::QueueInfoConteiner>>();
}

bool DapGetWalletInfoCommand::getAvailableBalance(const QString& walletName, const QString& datoshi, const QString& networkName, const QString& ticker,
                                                   QString& availableDatoshi, QString& availableCoins)
{
    auto feeInfo = getFee()[networkName];
    Dap::Coin res = dap_uint256_scan_uninteger(datoshi.toStdString().data());

    bool isHasDatoshi = false;

    for(const auto& item: m_queueData)
    {
        if(item.walletName == walletName && item.networkName == networkName && ticker == feeInfo.valTicker)
        {
            Dap::Coin value = dap_uint256_scan_uninteger(item.fee.toStdString().data());
            res = res - value;
            isHasDatoshi = true;
        }
        if(item.walletName == walletName && item.networkName == networkName && ticker == feeInfo.netTicker)
        {
            Dap::Coin value = feeInfo.netCoins;
            res = res - value;
            isHasDatoshi = true;
        }
        if(walletName == item.walletName && networkName == item.networkName && ticker == item.ticker)
        {
            Dap::Coin value = dap_uint256_scan_uninteger(item.value.toStdString().data());
            res = res - value;
            isHasDatoshi = true;
        }
    }
    if(!isHasDatoshi)
    {
        return false;
    }
    availableDatoshi = res.toDatoshiString();
    availableCoins = res.toCoinsString();
    return true;
}
