#include "DapVersionController.h"

const QString DapVersionController::LAST_VERSION = "lastVersion";
const QString DapVersionController::HAS_UPDATE = "hasUpdate";
const QString DapVersionController::MESSAGE = "message";
const QString DapVersionController::URL = "url";

DapVersionController::DapVersionController(const QString &asServicename, QObject *parent, const QString &cliPath)
    : DapAbstractCommand(asServicename, parent, cliPath)
{
    QString os_name;
#ifdef Q_OS_WIN
#define DOWNLOAD_PATH QString("C:/Users/Public/Documents/Cellframe-Dashboard/data/")
#define SCRIPT_PATH QString("C:/Users/Public/Documents/Cellframe-Dashboard/data/")
    os_name = "windows";
    s_suffix =  QString(".exe");

#elif defined Q_OS_LINUX
#define DOWNLOAD_PATH QString("/opt/")
#define SCRIPT_PATH QString("/opt/")
    os_name = "linux";
    s_suffix =  QString(".deb");

#elif defined Q_OS_MAC
#define DOWNLOAD_PATH QString("/Users/%1/Applications/").arg(getenv("USER"))
#define SCRIPT_PATH QString("/Users/%1/Applications/").arg(getenv("USER"))
    os_name = "macos";
    s_suffix =  QString(".pkg");

#elif defined Q_OS_ANDROID
    os_name = "android";
    s_suffix =  QString(".apk");
#endif
    if(DAP_BRAND_LO == "cellframe-wallet")
        s_urlPath = QString("https://pub.cellframe.net/%1/%2/main/latest/").arg(os_name).arg(DAP_BRAND_LO);
    else
        s_urlPath = QString("https://pub.cellframe.net/%1/%2/master/latest.amd64/").arg(os_name).arg(DAP_BRAND_LO);

    s_netManager = new QNetworkAccessManager(this);
    checkUpdate();
}


/// Send a response to the client.
/// @details Performed on the service side.
/// @param arg1...arg10 Parameters.
/// @return Reply to client.
QVariant DapVersionController::respondToClient(const QVariant &args)
{
    DapAbstractCommand::respondToClient(args);
    QStringList params = args.toStringList();
    bool isWeb3 = params.contains(WEB3_KEY);
    QJsonObject data;
    QJsonObject resultObj;
    QString errorMsg = "";

    if(isWeb3)
    {
        QString nodeVersion = getVersion();
        if(nodeVersion.isEmpty())
            resultObj.insert(ERROR_KEY, "Error. Can't get version of node");

        data.insert("cellframe-dashboard", DAP_VERSION);
        data.insert("cellframe-node", nodeVersion);

        resultObj.insert(RESULT_KEY, data);
        addWeb3Result(resultObj, args);
    }
    else
    {
        if(params.first() == "version")
        {
            checkUpdate();
            quint64 savedTime = QDateTime::currentSecsSinceEpoch();

            while(!s_flagWait){

                if(QDateTime::currentSecsSinceEpoch() - savedTime >= 10)
                {
                    qWarning()<<"Excessive waiting time for a response from the site";
                    s_error = "Excessive waiting time for a response from the site";
                    s_flagWait = true;
                }

                QCoreApplication::processEvents(QEventLoop::AllEvents);
            }
            if(s_error == "")
            {
                data.insert(LAST_VERSION,QJsonValue(s_numberVersion));
                data.insert(HAS_UPDATE,QJsonValue(s_hasUpdate));
                data.insert(MESSAGE,QJsonValue("Reply version"));
                data.insert(URL,QJsonValue(QString(s_urlPath + "../" + s_lastVersion.toString())));
            }
            else
            {
                errorMsg = "Error. " + s_error;
                data.insert(MESSAGE,QJsonValue(errorMsg));
            }
        }
        else if(params.first() == "version node")
        {
            QString nodeVersion = getVersion();
            if(nodeVersion.isEmpty())
                errorMsg = "Error. Can't get version of node";

            if(!errorMsg.isEmpty())
            {
                data.insert(MESSAGE,QJsonValue(errorMsg));
            }
            else
            {
                data.insert(LAST_VERSION,QJsonValue(nodeVersion));
                data.insert(MESSAGE,QJsonValue("Reply node version"));
            }
        }

        if(!errorMsg.isEmpty())
            resultObj.insert(ERROR_KEY, errorMsg);
        resultObj.insert(RESULT_KEY, data);
    }

    QJsonDocument resultDoc(resultObj);
    return resultDoc.toJson();
}

void DapVersionController::replyFromService()
{
    DapRpcServiceReply *reply = static_cast<DapRpcServiceReply *>(sender());
    QVariant replyVariant = reply->response().toJsonValue().toVariant();

    QJsonDocument replyDoc = QJsonDocument::fromJson(replyVariant.toByteArray());
    QJsonObject replyObj = replyDoc.object();
    QJsonObject resultObj = replyObj[RESULT_KEY].toObject();

    emit serviceResponded(resultObj);
}

void DapVersionController::checkUpdate()
{
    s_flagWait = false;
    QPointer<QNetworkAccessManager> manager = new QNetworkAccessManager(this);
    QNetworkRequest request((QUrl(s_urlPath)));
    qDebug() << "[Test_build] [DapVersionController] tryRequest";

    connect(manager, &QNetworkAccessManager::finished, [this, manager](QNetworkReply *reply) {

        s_error = "";
        qDebug() << "[Test_build] [DapVersionController] rcv";
        if (reply->error() == QNetworkReply::NoError ||
            reply->error() == QNetworkReply::ContentNotFoundError)
            {
            QByteArray content= reply->readAll();
            QTextCodec *codec = QTextCodec::codecForName("utf8");
            QString str = codec->toUnicode(content.data());
            QRegExp rw("[\\w+\\s+-\\.]+" + s_suffix);
            QString fileName;
            rw.indexIn(str);
            fileName = rw.cap(0);

            if(fileName != ""){
                s_lastVersion = fileName;
                s_hasUpdate = compareVersion();
                //            installPack(); FOR DEBUG WIN
            }
        }else{
            qWarning()<<reply->errorString();
            s_error = reply->errorString();
        }
        qDebug() << "[Test_build] [DapVersionController] rcv delete";
        reply->deleteLater();
        manager->deleteLater();
        if(!s_flagWait)
            s_flagWait = true;
    });

    QNetworkReply *rep = manager->get(request);
//    qDebug()<<rep->errorString();
}

bool DapVersionController::compareVersion()
{
    QString currVersion = DAP_VERSION;

    s_numberVersion = s_lastVersion.toString().split("_").at(1).split("_").at(0);
    QStringList ver = s_numberVersion.split(".");
    s_numberVersion.replace(s_numberVersion.lastIndexOf('.'), 1, '-');

    //Test operator >
    //    return s_numberVersion > currVersion ? true : false;
    //    int res = QString::compare(s_numberVersion, currVersion);

    int lastMaj = ver[0].toInt();
    int currMaj = currVersion.split(".").at(0).toInt();

    int lastMin = ver[1].toInt();
    int currMin = currVersion.split(".").at(1).split("-").at(0).toInt();

    int lastPat = ver[2].toInt();
    int currPat = currVersion.split(".").at(1).split("-").at(1).toInt();

    qDebug()<<"Last " << lastMaj << "." << lastMin << "-" <<lastPat << " ---- " <<
        "Curr " << currMaj << "." << currMin << "-" <<currPat;

    if(lastMaj > currMaj)
        return true;
    else if(lastMaj < currMaj)
        return false;

    if(lastMin > currMin)
        return true;
    else if(lastMin < currMin)
        return false;

    if(lastPat > currPat)
        return true;
    else if(lastPat < currPat)
        return false;

    return false;
}

bool DapVersionController::installPack()
{
#ifdef Q_OS_LINUX
    createInstallScript();
    system("chmod 777 /opt/installScript.sh");
    system("/opt/installScript.sh &");
#elif defined Q_OS_WIN
    QString name = s_lastVersion.toString();
    QString path = DOWNLOAD_PATH;
    path += name;
    s_downloadFilePath = path;

    QNetworkRequest request;
    request.setUrl(QUrl(s_urlPath + "../" + name));

    s_downloadfile = new QFile(s_downloadFilePath);
    if(s_downloadfile->exists())
    {
        s_downloadfile->remove();
        s_downloadfile->deleteLater();
        s_downloadfile = new QFile(s_downloadFilePath);
    }
    s_downloadfile->open(QIODevice::ReadWrite | QIODevice::Append);
    QNetworkReply *reply = s_netManager->get(request);
    connect(reply, &QNetworkReply::finished,this, &DapVersionController::downloadPack);

#endif

    return true;
}


void DapVersionController::downloadPack()
{
    QNetworkReply *reply = qobject_cast<QNetworkReply *>(sender());
    QVariant statusCode = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute);
    QString path = DOWNLOAD_PATH;
    QString pathPack = path + s_lastVersion.toString();

    if (reply->error() == QNetworkReply::NoError ||
        reply->error() == QNetworkReply::ContentNotFoundError)
    {
        s_downloadfile->write(reply->readAll());
        s_downloadfile->close();
    }
    else
        qWarning()<<"Download error";

    qInfo()<<"Reply finished. Status code: " << statusCode.toInt();

    reply->deleteLater();
    s_downloadfile->deleteLater();

    qInfo() << "Udpating dashboard: "
            << QProcess::startDetached("schtasks.exe",
                                       QStringList({"/run", "/TN", pathPack}));
}

void DapVersionController::createInstallScript()
{
    QString path = SCRIPT_PATH + "installScript.sh";
    s_installScript = new QFile(path);

    if(s_installScript->exists())
    {
        s_installScript->remove();
        s_installScript->deleteLater();
        s_installScript = new QFile(path);
    }
    if(s_installScript->open(QIODevice::ReadWrite | QIODevice::Append))
    {
        QTextStream out(s_installScript);
        out <<" tty -s || exec x-terminal-emulator -e /opt/installScript.sh" << endl;
        out <<"  I=$( wget -O cellframe-dashboard.deb "+s_urlPath + "../" + s_lastVersion.toString()+ ")"<< endl;
        out <<" P='cellframe-dashboard.deb'" << endl;
        out <<" sudo dpkg -i $P" << endl;
        out <<"J=`dpkg -s cellframe-dashboard | grep \"/Status\"/ `"<< endl;
        out <<"if [ -n \"/$J\"/ ]"<< endl;
        out <<"then" << endl;
        out <<"    echo \"/==============================================================\"/"<< endl;
        out <<"    echo                            $J                                  "<< endl;
        out <<"    echo \"/==============================================================\"/"<< endl;
        out <<"else"<< endl;
        out <<"    echo \"/==============================================================\"/"<< endl;
        out <<"    echo                            $J                                  "<< endl;
        out <<"    echo \"/==============================================================\"/"<< endl;
        out <<"fi"<< endl;
        out <<"rm $P"<< endl;
        out <<"cd /opt/cellframe-dashboard/bin/"<<endl;
        out <<"nohup  ./Cellframe-Dashboard"<<endl;
        s_installScript->close();
    }
}
