#include "DapCommandStackManager.h"
#include "TransactionQueue/DapTransactionQueueController.h"

#include <dap_hash.h>
#include <QRandomGenerator>

DapCommandStackManager::DapCommandStackManager(const DapAbstractCommand *command)
    : QObject()
{
    m_nameCommand = command->getName();
    DapTransactionQueueController* controller = DapTransactionQueueController::getTransactionController();
    connect(controller, &DapTransactionQueueController::toCreateTransaction, [this](const QString& nameCommand, const QStringList& paramCommand)
            {
                if(this->m_nameCommand == nameCommand)
                {
                    QString problem = "There are problems with " + m_nameCommand;
                    Q_ASSERT_X(m_commandCallback, "DapCommandStackManager", problem.toStdString().c_str());
                    QStringList withNewParam = paramCommand;
                    qDebug() << "[Queue] [DapCommandStackManager] Trying to create a transaction. Command: " << nameCommand << ", param list: " << paramCommand;
                    withNewParam.append("self");
                    m_commandCallback(std::move(withNewParam));
                }
            });
}

QVariant DapCommandStackManager::respondToClient(const QVariant &args)
{
    QJsonObject error;
    QJsonObject resultObj;
    bool isQueue = false;
    QString hash, queueHash;
    QJsonValue successValue;
    DapTransactionQueueController* controller = DapTransactionQueueController::getTransactionController();
    if(!controller)
    {
        qWarning() << "[Queue] DapTransactionQueueController not initialized";
        QString problem = "There are problems with " + m_nameCommand;
        Q_ASSERT_X(m_parentCallback, "DapCommandStackManager", problem.toStdString().c_str());
        return m_parentCallback(args);
    }
    else
    {
        QStringList listArgument = args.toStringList();
        if(!listArgument.contains(HASH_QUEUE_KEY))
        {
            listArgument.append(HASH_QUEUE_KEY);
            queueHash = getNewId();
            listArgument.append(queueHash);
        }
        if(listArgument.isEmpty())
        {
            qWarning() << "[Queue] No arguments";
        }

        bool isSelf = listArgument.back() == "self";

        QString problem = "There are problems with networkNameCallback " + m_nameCommand;
        Q_ASSERT_X(m_networkNameCallback, "DapCommandStackManager", problem.toStdString().c_str());
        QString networkName = m_networkNameCallback(listArgument);//[0];

        problem = "There are problems with walletNameCallback " + m_nameCommand;
        Q_ASSERT_X(m_walletNameCallback, "DapCommandStackManager", problem.toStdString().c_str());
        QString walletName = m_walletNameCallback(listArgument);//[1];

//        QStringList argsList = args.toStringList();

        if(controller->isReady(networkName, walletName))
        {
            QString commandName = m_nameCommand;

            QString problem = "There are problems with parentCallback " + m_nameCommand;
            Q_ASSERT_X(m_parentCallback, "DapCommandStackManager", problem.toStdString().c_str());
            QVariant result = m_parentCallback(listArgument);
            QJsonDocument replyDoc = QJsonDocument::fromJson(result.toByteArray());
            QJsonObject replyObj = replyDoc.object();

            if(replyObj.contains("error"))
            {
                if(replyObj.value("error").isString())
                {
                    error.insert("message", replyObj.value("error").toString());
                }
                else
                {
                    error = replyObj.value("error").toObject();
                }
            }

            QJsonObject object = replyObj["result"].toObject();
            successValue = object["success"];
            if(object.contains("tx_hash"))
            {
                hash = object["tx_hash"].toString();
                for(const QString& key: object.keys())
                {
                    resultObj.insert(key, object.value(key));
                }
                // resultObj.insert("tx_hash", hash);

                qInfo() << QString("[Queue] The transaction got into the mempool and received a hash of %1").arg(hash);

                if(isSelf)
                {
                    qDebug() << "[Queue] [DapCreateTransactionCommandStack] A transaction has been created from the queue";
                    controller->addTransactionHash(networkName, walletName, hash);
                }
                else
                {
                    qDebug() << "[Queue] [DapCreateTransactionCommandStack] A new transaction has been created and added to the queue for verification";

                    controller->addNewTransaction(m_nameCommand, listArgument, networkName, walletName, hash, queueHash);
                    isQueue = true;
                }
            }
            else if(!successValue.toBool())
            {
                if(isSelf)
                {
                    controller->clearCurrentTransaction(networkName, walletName);
                }
                if(error.isEmpty())
                {
                    error.insert("message", object["message"].toString());
                }
            }
        }
        else if(!isSelf)
        {
            qDebug() << "[Queue] [DapCreateTransactionCommandStack] Added a transaction to the queue";
            controller->addNewTransaction(m_nameCommand, listArgument, networkName, walletName, "", queueHash);
            isQueue = true;
        }
    }
    bool toQueue = hash.isEmpty() ? isQueue : false;

    if(resultObj.contains("tx_hash"))
    {
        if(!resultObj["tx_hash"].toString().contains(QRegularExpression(R"(0x[a-fA-F0-9])")))
        {
            error.insert("message", "An invalid hash value was received. Perhaps the answer has changed in the node.");
        }
    }

    resultObj.insert("idQueue", queueHash);
    resultObj.insert("toQueue", toQueue);
    resultObj.insert("success", isQueue ? true : successValue);

    QJsonDocument resultDoc;
    QJsonObject mainObject;
    if(!error.isEmpty())
    {
        mainObject.insert("error", error);
    }

    mainObject.insert("result", resultObj);
    if(m_addWeb3InfoCallBack)
    {
        m_addWeb3InfoCallBack(mainObject, args);
    }
    resultDoc.setObject(mainObject);

    return resultDoc.toJson();
}

QString DapCommandStackManager::getRandomString()
{
    QString possibleCharacters("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789");
    int randomStringLength = 8;
    QString randomString;
    for(int i=0; i<randomStringLength; ++i)
    {
        int index = QRandomGenerator::global()->bounded(0, possibleCharacters.length());
        QChar nextChar = possibleCharacters.at(index);
        randomString.append(nextChar);
    }
    return randomString;
}

QString DapCommandStackManager::getNewId()
{
    dap_chain_hash_fast_t l_hash_cert_pkey;
    QString id = getRandomString();
    dap_hash_fast(id.constData(), id.size(), &l_hash_cert_pkey);
    char *l_cert_pkey_hash_str = dap_chain_hash_fast_to_str_new(&l_hash_cert_pkey);
    QString result = QString::fromLatin1(l_cert_pkey_hash_str);
    DAP_DEL_Z(l_cert_pkey_hash_str)
    return result;
}
