#include "DapCreateJsonTransactionCommandStack.h"
#include "TransactionQueue/DapTransactionQueueController.h"
#include <QFile>

DapCreateJsonTransactionCommandStack::DapCreateJsonTransactionCommandStack(const QString &asServicename, QObject *parent, const QString &asCliPath)
    : DapCreateJsonTransactionCommand(asServicename, parent, asCliPath)
    , m_commandManager(new DapCommandStackManager(this))
{
    m_commandManager->setCommandCalback([&](const QVariant &args) -> QVariant { return this->respondToClient(args);});
    m_commandManager->setParentCalback([&](const QVariant &args) -> QVariant { return DapCreateJsonTransactionCommand::respondToClient(args); });
    m_commandManager->setWalletNameCalback([&](const QStringList &list) -> QString { return getData(list, DataType::WALLET_NAME); });
    m_commandManager->setNetworkNameCalback([&](const QStringList &list) -> QString { return getData(list, DataType::NET);});
    m_commandManager->setWeb3UpdateCalback([&](QJsonObject& result, const QVariant& args){ addWeb3Result(result, args); });

    DapTransactionQueueController* controller = DapTransactionQueueController::getTransactionController();
    DapTransactionQueueController::ParamsFromCommand callbacks;
    callbacks.addressCallback       = [&](const QStringList &list) -> QString { return getData(list, DataType::ADDRESS);};
    callbacks.feeCallback           = [&](const QStringList &list) -> QString { return getData(list, DataType::FEE);};
    callbacks.tokenCallback         = [&](const QStringList &list) -> QString { return getData(list, DataType::TOKEN);};
    callbacks.valueCallback         = [&](const QStringList &list) -> QString { return getData(list, DataType::VALUE);};
    callbacks.walletNameCallback    = [&](const QStringList &list) -> QString { return getData(list, DataType::WALLET_NAME);};
    callbacks.directionCallback     = [](const QStringList &list) -> QString { return "to";};
    callbacks.testExpenses          = []() -> bool { return false;};
    callbacks.feeInfo               = [&]() -> const QMap<QString, NetworkSpace::FeeInfoStruct> { return getFee();};
    controller->setParamsCommandCallback(this->getName(), callbacks);
    controller->addCommandWithoutInfoWallet(this->getName());
}

QVariant DapCreateJsonTransactionCommandStack::respondToClient(const QVariant &args)
{
    QStringList listData = args.toStringList();

    QRegularExpressionMatch match = REGULAR_JSON.match(listData.at(0));

    if (!match.hasMatch())
    {
        qWarning() << "[DapCreateJsonTransactionCommandStack] Incorrect json data";

        QJsonDocument resultDoc;
        QJsonObject resultObj;
        resultObj.insert("success", false);
        resultObj.insert("error", "Incorrect json data");

        resultObj.insert("result", resultObj);
        addWeb3Result(resultObj, args);

        resultDoc.setObject(resultObj);

        return resultDoc.toJson();
    }
    return m_commandManager->respondToClient(args);
}

DapCreateJsonTransactionCommandStack::~DapCreateJsonTransactionCommandStack()
{
    delete m_commandManager;
}

QString DapCreateJsonTransactionCommandStack::getData(const QStringList& list, DataType type)
{
    QRegularExpressionMatch match = REGULAR_JSON.match(list.at(0));
    if (!match.hasMatch())
    {
        qWarning() << "[DapCreateJsonTransactionCommandStack][getData] Incorrect json data";
        return QString();
    }
    QJsonDocument jsonCommand = QJsonDocument::fromJson(match.captured(0).toStdString().data());
    QJsonObject dataObject = jsonCommand.object();

    auto getItem = [&](DataType type) ->QString
    {
        QJsonArray items = dataObject["items"].toArray();
        auto feeInfo = getFee()[dataObject["net"].toString()];
        for(const auto& value: items)
        {
            QJsonObject item = value.toObject();
            switch (type)
            {
            case DataType::FEE:
            {
                if(item.contains("subtype") && item.take("subtype").toString() == "fee")
                {
                    return item["value"].toString();
                }
            }
            break;
            case DataType::WALLET_NAME:
            {
                if(item.contains("type") && item.take("type").toString() == "sign")
                {
                    return item["wallet"].toString();
                }
            }
                break;
            case DataType::ADDRESS:
            case DataType::TOKEN:
            case DataType::VALUE:
            {
                if(!item.contains("type") || !item.take("type").toString().contains("out"))
                {
                    continue;
                }
                if(!item.contains("addr"))
                {
                    continue;
                }
                QString addr = item["addr"].toString();
                if(addr == feeInfo.netAddr)
                {
                    continue;
                }
                switch (type)
                {
                case DataType::ADDRESS:
                    return addr;
                    break;
                case DataType::TOKEN:
                    return item["token"].toString();
                    break;
                case DataType::VALUE:
                    return item["value"].toString();
                    break;
                }
            }
                break;
            }
        }
        return QString();
    };

    switch (type)
    {
    case DataType::NET:
        return dataObject["net"].toString();
        break;
    default:
        return getItem(type);
        break;
    }

    return QString();
}
